;;;			     tui-disp.el
;;;
;;;	      Copyright (C) 1997       OKUJI Yoshinori
;;;
;;;		   This emacs lisp library conforms
;;;		GNU GENERAL PUBLIC LICENSE Version 2.
;;;
;;; Author:  OKUJI Yoshinori <okuji@kmc.kyoto-u.ac.jp>
;;; Created: May    15, 1997
;;; Revised: May    26, 1997
;;;

;;;
;;; Loading.
;;;
(provide 'tui-disp)
(require 'tuitate)

;;;
;;; Constant variables.
;;;
(defconst tuitate-disp-right-belongings-col   48)
(defconst tuitate-disp-right-turn-col         49)
(defconst tuitate-disp-name-line              1)
(defconst tuitate-disp-right-name-col         49)
(defconst tuitate-disp-life-line              2)
(defconst tuitate-disp-right-life-col         49)
(defconst tuitate-disp-board-line             3)
(defconst tuitate-disp-belongings-line        9)
(defconst tuitate-disp-belonging-format       "  %c  %2d ")
(defconst tuitate-disp-belonging-width        8)
(defconst tuitate-disp-belonging-piece-offset 1)
(defconst tuitate-disp-promoted-pieces
  '(Pawn Lance Knight Silver Bishop Rook))
(defconst tuitate-disp-right-piece-chars
  '((pawn   . ?p)
    (lance  . ?l)
    (knight . ?n)
    (silver . ?s)
    (gold   . ?g)
    (bishop . ?b)
    (rook   . ?r)))
(defconst tuitate-disp-left-piece-chars
  '((pawn   . ?P)
    (lance  . ?L)
    (knight . ?N)
    (silver . ?S)
    (gold   . ?G)
    (bishop . ?B)
    (rook   . ?R)))
(defconst tuitate-disp-belongings-tab         2)
(defconst tuitate-disp-board-origin-line      5)
(defconst tuitate-disp-board-origin-col       12)
(defconst tuitate-disp-square-x-size          4)
(defconst tuitate-disp-square-y-size          2)
(defconst tuitate-disp-piece-width            3)
(defconst tuitate-disp-right-piece-strings
  '((nil    . "   ")
    (pawn   . " p ")
    (lance  . " l ")
    (knight . " n ")
    (silver . " s ")
    (gold   . " g ")
    (king   . " k ")
    (bishop . " b ")
    (rook   . " r ")
    (Pawn   . "*p*")
    (Lance  . "*l*")
    (Knight . "*n*")
    (Silver . "*s*")
    (Bishop . "*b*")
    (Rook   . "*r*")))
(defconst tuitate-disp-left-piece-strings
  '((nil    . "   ")
    (pawn   . " P ")
    (lance  . " L ")
    (knight . " N ")
    (silver . " S ")
    (gold   . " G ")
    (king   . " K ")
    (bishop . " B ")
    (rook   . " R ")
    (Pawn   . "*P*")
    (Lance  . "*L*")
    (Knight . "*N*")
    (Silver . "*S*")
    (Bishop . "*B*")
    (Rook   . "*R*")))

(defconst tuitate-disp-board-string
  "             9   8   7   6   5   4   3   2   1  
           +---+---+---+---+---+---+---+---+---+
         a |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         b |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         c |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         d |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         e |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         f |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         g |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         h |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         i |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+"
  )

(defconst tuitate-disp-reverse-board-string
  "             1   2   3   4   5   6   7   8   9  
           +---+---+---+---+---+---+---+---+---+
         i |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         h |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         g |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         f |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         e |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         d |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         c |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         b |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+
         a |   |   |   |   |   |   |   |   |   |
           +---+---+---+---+---+---+---+---+---+"
  )

;;;
;;; Faces.
;;;
(defun tuitate-disp-set-face (face fg bg)
  (set-face-foreground face fg)
  (set-face-background face bg))

(defun tuitate-disp-make-faces ()
  (make-face 'tuitate-info-face)	; turn, name, and life.
  (make-face 'tuitate-invert-info-face)	; if current turn, turn inverted.
  (make-face 'tuitate-board-face)	; board.
  (make-face 'tuitate-right-piece-face)	; right pieces.
  (make-face 'tuitate-left-piece-face)	; left pieces.
  (make-face 'tuitate-right-promoted-piece-face) ; right promoted pieces.
  (make-face 'tuitate-left-promoted-piece-face) ; left promoted pieces.
  (make-face 'tuitate-belongings-face)	; belongings.
  (make-face 'tuitate-hold-face)	; selected piece.
  (make-face 'tuitate-candidate-face)	; candidates for destination.

  (cond ((and (not tuitate-inhibit-color)
	      (eq window-system 'x))
	 ;; colorful settings.
	 (tuitate-disp-set-face 'tuitate-info-face
				"sky blue"
				"dark green")
	 (tuitate-disp-set-face 'tuitate-invert-info-face
				"dark green"
				"sky blue")
	 (tuitate-disp-set-face 'tuitate-board-face
				"black"
				"dark orange")
	 (tuitate-disp-set-face 'tuitate-right-piece-face
				"sea green"
				"dark orange")
	 (tuitate-disp-set-face 'tuitate-left-piece-face
				"dark red"
				"dark orange")
	 (tuitate-disp-set-face 'tuitate-right-promoted-piece-face
				"green"
				"dark orange")
	 (tuitate-disp-set-face 'tuitate-left-promoted-piece-face
				"red"
				"dark orange")
	 (tuitate-disp-set-face 'tuitate-belongings-face
				"black"
				"dark orange")
	 (tuitate-disp-set-face 'tuitate-hold-face
				"purple"
				"orange")
	 (tuitate-disp-set-face 'tuitate-candidate-face
				"blue"
				"grey"))
	((or (eq window-system 'x)
	     (not tuitate-inhibit-terminal-face))
	 ;; mono setting. Do you hate the colorful world?
	 (copy-face 'default 'tuitate-info-face)
	 (copy-face 'default 'tuitate-invert-info-face)
	 (invert-face 'tuitate-invert-info-face)
	 (copy-face 'default 'tuitate-board-face)
	 (copy-face 'default 'tuitate-right-piece-face)
	 (copy-face 'default 'tuitate-left-piece-face)
	 (copy-face 'default 'tuitate-right-promoted-piece-face)
	 (copy-face 'default 'tuitate-left-promoted-piece-face)
	 (copy-face 'default 'tuitate-belongings-face)
	 (copy-face 'default 'tuitate-hold-face)
	 (invert-face 'tuitate-hold-face)
	 (copy-face 'default 'tuitate-candidate-face)
	 (invert-face 'tuitate-candidate-face))))

;;;
;;; Display subroutines.
;;;

;; A data length and its looks is not same on Mule.
(defalias 'tuitate-disp-string-length 
  (if (boundp 'MULE)
      'string-width
    'length))

;; If the buffer is too small, add lines automatically.
(defun tuitate-disp-goto-line (line)
  (let ((rest (goto-line line))
	(pnt (point)))
    (beginning-of-line)
    (if (not (= (point) pnt))
	(progn
	  (setq rest 1)
	  (end-of-line)))
    (if (> rest 0)
	(insert (make-string rest ?\n)))))

;; Insert "White ... Black", normally.
;;(defun tuitate-disp-insert-turn-string (&optional reverse)
;;  (save-excursion
;;    (set-buffer tuitate-buffer-name)
;;    (let ((buffer-read-only nil))
;;      (tuitate-disp-goto-line tuitate-disp-turn-line)
;;      (let ((left (if reverse tuitate-black-string
;;		    tuitate-white-string))
;;	    (right (if reverse tuitate-white-string
;;		     tuitate-black-string)))
;;	(if (or (eq window-system 'x)
;;		(not tuitate-inhibit-terminal-face))
;;	    (progn
;;	      (put-text-property 0
;;				 (length left)
;;				 'face
;;				 'tuitate-info-face
;;				 left)
;;	      (put-text-property 0
;;				 (length right)
;;				 'face
;;				 'tuitate-info-face
;;				 right)))
;;	(insert left
;;		(make-string (- tuitate-disp-right-turn-col
;;				(tuitate-disp-string-length left))
;;			     ?\ )
;;		right)))))

;; Insert "Opponent ... You", and so on.
(defun tuitate-disp-insert-name-string (left-name right-name &optional game)
  (save-excursion
    (set-buffer tuitate-buffer-name)
    (let ((buffer-read-only nil))
      (tuitate-disp-goto-line tuitate-disp-name-line)
      (let ((left (concat (if game tuitate-opponent-string tuitate-name-string)
			  ": "
			  left-name))
	    (right (concat (if game tuitate-you-string tuitate-name-string)
			   ": "
			   right-name)))
	(if (or (eq window-system 'x)
		(not tuitate-inhibit-terminal-face))
	    (progn
	      (put-text-property 0
				 (length left)
				 'face
				 'tuitate-info-face
				 left)
	      (put-text-property 0
				 (length right)
				 'face
				 'tuitate-info-face
				 right)))
	(insert left
		(make-string (- tuitate-disp-right-name-col
				(tuitate-disp-string-length left))
			     ?\ )
		right)))))

;; Overwrite "Life: ? ... Life: ?".
(defun tuitate-disp-write-life-string (left-life right-life)
  (save-excursion
    (set-buffer tuitate-buffer-name)
    (let ((buffer-read-only nil))
      (tuitate-disp-goto-line tuitate-disp-life-line)
      (let ((old (point)))
	(end-of-line)
	(or (= old (point)) (delete-region old (point))))
      (let ((left (concat tuitate-life-string
			  ": "
			  (number-to-string left-life)))
	    (right (concat tuitate-life-string
			   ": "
			   (number-to-string right-life))))
	(if (or (eq window-system 'x)
		(not tuitate-inhibit-terminal-face))
	    (progn
	      (put-text-property 0
				 (length left)
				 'face
				 'tuitate-info-face
				 left)
	      (put-text-property 0
				 (length right)
				 'face
				 'tuitate-info-face
				 right)))
	(insert left
		(make-string (- tuitate-disp-right-life-col
				(tuitate-disp-string-length left))
			     ?\ )
		right)))))

;; Insert the entire board.
(defun tuitate-disp-insert-board (&optional reverse)
  (save-excursion
    (set-buffer tuitate-buffer-name)
    (let ((buffer-read-only nil))
      (tuitate-disp-goto-line tuitate-disp-board-line)
      (if reverse
	  (insert tuitate-disp-reverse-board-string)
	(insert tuitate-disp-board-string))
      (let ((i 0) start end)
	(goto-line (1- tuitate-disp-board-origin-line))
	(while (< i 19)
	  (forward-char (1- tuitate-disp-board-origin-col))
	  (setq start (point))
	  (end-of-line)
	  (setq end (point))
	  (if (or (eq window-system 'x)
		  (not tuitate-inhibit-terminal-face))
	      (put-text-property start end 'face 'tuitate-board-face))
	  (forward-line 1)
	  (setq i (1+ i)))))))

;; Overwrite players' belongings.
(defun tuitate-disp-write-belongings (left-alist right-alist)
  (save-excursion
    (set-buffer tuitate-buffer-name)
    (let ((buffer-read-only nil))
      (if left-alist
	  (progn
	    (goto-line tuitate-disp-belongings-line)
	    (mapcar (function (lambda (elm)
				(let ((ins
				       (format
					tuitate-disp-belonging-format
					(cdr (assq
					      (car elm)
					      tuitate-disp-left-piece-chars))
					(cdr elm))))
				  (delete-char tuitate-disp-belonging-width)
				  (if (or (eq window-system 'x)
					  (not tuitate-inhibit-terminal-face))
				      (put-text-property
				       1
				       tuitate-disp-belonging-width
				       'face
				       'tuitate-belongings-face
				       ins))
				  (insert ins)
				  (forward-line 1)
				  (delete-char tuitate-disp-belonging-width)
				  (setq ins
					(make-string
					 tuitate-disp-belonging-width ?\ ))
				  (if (or (eq window-system 'x)
					  (not tuitate-inhibit-terminal-face))
				      (put-text-property
				       1
				       tuitate-disp-belonging-width
				       'face
				       'tuitate-belongings-face
				       ins))
				  (insert ins)
				  (forward-line 1))))
		    left-alist)))
      (if right-alist
	  (progn
	    (goto-line tuitate-disp-belongings-line)
	    (mapcar (function (lambda (elm)
				(forward-char
				 tuitate-disp-right-belongings-col)
				(let ((old (point)))
				  (end-of-line)
				  (or (= (point) old)
				      (delete-region old (point))))
				(let ((ins (format
					    tuitate-disp-belonging-format
					    (cdr
					     (assq
					      (car elm)
					      tuitate-disp-right-piece-chars))
					    (cdr elm))))
				  (if (or (eq window-system 'x)
					  (not tuitate-inhibit-terminal-face))
				      (put-text-property
				       1
				       tuitate-disp-belonging-width
				       'face
				       'tuitate-belongings-face
				       ins))
				  (insert ins)
				  (forward-line 1)
				  (forward-char
				   tuitate-disp-right-belongings-col)
				  (let ((old (point)))
				    (end-of-line)
				    (or (= (point) old)
					(delete-region old (point))))
				  (setq ins
					(make-string
					 tuitate-disp-belonging-width ?\ ))
				  (if (or (eq window-system 'x)
					  (not tuitate-inhibit-terminal-face))
				      (put-text-property
				       1
				       tuitate-disp-belonging-width
				       'face
				       'tuitate-belongings-face
				       ins))
				  (insert ins)
				  (forward-line 1))))
		    right-alist))))))

;; Overwrite pieces on the board.
(defun tuitate-disp-write-piece (piece x y &optional left)
  (save-excursion
    (set-buffer tuitate-buffer-name)
    (let ((buffer-read-only nil))
      (goto-line (+ tuitate-disp-board-origin-line
		    (* y tuitate-disp-square-y-size)))
      (forward-char (+ tuitate-disp-board-origin-col
		       (* x tuitate-disp-square-x-size)))
      (delete-char tuitate-disp-piece-width)
      (let ((ins (if left
		     (cdr (assq piece tuitate-disp-left-piece-strings))
		   (cdr (assq piece tuitate-disp-right-piece-strings)))))
	(if (or (eq window-system 'x)
		(not tuitate-inhibit-terminal-face))
	    (if (memq piece tuitate-disp-promoted-pieces)
		;; promoted piece.
		(if left
		    (put-text-property 0
				       tuitate-disp-piece-width
				       'face
				       'tuitate-left-promoted-piece-face
				       ins)
		  (put-text-property 0
				     tuitate-disp-piece-width
				     'face
				     'tuitate-right-promoted-piece-face
				     ins))
	      ;; non-promoted piece.
	      (if left
		  (put-text-property 0
				     tuitate-disp-piece-width
				     'face
				     'tuitate-left-piece-face
				     ins)
		(put-text-property 0
				   tuitate-disp-piece-width
				   'face
				   'tuitate-right-piece-face
				   ins))))
	(insert ins)))))

;; When you hold a piece, it is highlighted.
(defun tuitate-disp-hold-piece (x y &optional delete)
  (if (or (eq window-system 'x)
	  (not tuitate-inhibit-terminal-face))
      (save-excursion
	(set-buffer tuitate-buffer-name)
	(let ((buffer-read-only nil))
	  (goto-line (+ tuitate-disp-board-origin-line
			(* y tuitate-disp-square-y-size)))
	  (forward-char (+ tuitate-disp-board-origin-col
			   (* x tuitate-disp-square-x-size)))
	  (let* ((start (point))
		 (end (+ start tuitate-disp-piece-width))
		 ol)
	    (if delete
		(progn
		  (setq ol (overlays-at start))
		  (mapcar (function delete-overlay) ol))
	      (setq ol (make-overlay start end))
	      (overlay-put ol 'face 'tuitate-hold-face)))))))

;; When you hold a belonging, it is highlighted.
(defun tuitate-disp-hold-belonging (index &optional delete left)
  (if (or (eq window-system 'x)
	  (not tuitate-inhibit-terminal-face))
      (save-excursion
	(set-buffer tuitate-buffer-name)
	(let ((buffer-read-only nil))
	  (goto-line (+ tuitate-disp-belongings-line
			(* index tuitate-disp-belongings-tab)))
	  (if (not left)
	      (forward-char tuitate-disp-right-belongings-col))
	  (let* ((start (+ tuitate-disp-belonging-piece-offset (point)))
		 (end (+ start tuitate-disp-piece-width))
		 ol)
	    (if delete
		(progn
		  (setq ol (overlays-at start))
		  (mapcar (function delete-overlay) ol))
	      (setq ol (make-overlay start end))
	      (overlay-put ol 'face 'tuitate-hold-face)))))))

;; When a square is possible to put holding piece, it is highlighted.
(defun tuitate-disp-candidate (x y &optional delete)
  (if (or (eq window-system 'x)
	  (not tuitate-inhibit-terminal-face))
      (save-excursion
	(set-buffer tuitate-buffer-name)
	(let ((buffer-read-only nil))
	  (goto-line (+ tuitate-disp-board-origin-line
			(* y tuitate-disp-square-y-size)))
	  (forward-char (+ tuitate-disp-board-origin-col
			   (* x tuitate-disp-square-x-size)))
	  (let* ((start (point))
		 (end (+ start tuitate-disp-piece-width))
		 ol)
	    (if delete
		(progn
		  (setq ol (overlays-at start))
		  (mapcar (function delete-overlay) ol))
	      (setq ol (make-overlay start end))
	      (overlay-put ol 'face 'tuitate-candidate-face)))))))

;; Now which player is in turn? Oh, I see it easily!
(defun tuitate-disp-turn (left &optional reverse delete game)
  (if (or (eq window-system 'x)
	  (not tuitate-inhibit-terminal-face))
      (save-excursion
	(set-buffer tuitate-buffer-name)
	(let ((buffer-read-only nil))
	  (goto-line tuitate-disp-name-line)
	  (if (not left)
	      (let ((str (if game
			     tuitate-opponent-string
			   tuitate-name-string)))
		(forward-char (+ (- tuitate-disp-right-name-col
				    (length str))
				 (tuitate-disp-string-length str)))))
	  (let* ((str (if game
			  (if left tuitate-opponent-string tuitate-you-string)
			tuitate-name-string))
		 (start (point))
		 (end (+ start (length str)))
		 ol)
	    (if delete
		(progn
		  (setq ol (overlays-at start))
		  (mapcar (function delete-overlay) ol))
	      (setq ol (make-overlay start end))
	      (overlay-put ol 'face 'tuitate-invert-info-face)))))))

;;;
;;; Return the buffer point.
;;;
(defun tuitate-disp-piece-point (x y)
  (save-excursion
    (set-buffer tuitate-buffer-name)
    (goto-line (+ tuitate-disp-board-origin-line
		  (* y tuitate-disp-square-y-size)))
    (forward-char (+ tuitate-disp-board-origin-col
		     (* x tuitate-disp-square-x-size)))
    (1+ (point))))

(defun tuitate-disp-belonging-point (index &optional left)
  (save-excursion
    (set-buffer tuitate-buffer-name)
    (goto-line (+ tuitate-disp-belongings-line
		  (* index tuitate-disp-belongings-tab)))
    (if (not left)
	(forward-char tuitate-disp-right-belongings-col))
    (+ (point) tuitate-disp-belonging-piece-offset 1)))

;;;
;;; End of tui-disp.el
;;;
