;;;			     tui-game.el
;;;
;;;	      Copyright (C) 1997       OKUJI Yoshinori
;;;
;;;		   This emacs lisp library conforms
;;;		GNU GENERAL PUBLIC LICENSE Version 2.
;;;
;;; Author:  OKUJI Yoshinori <okuji@kmc.kyoto-u.ac.jp>
;;; Created: May    15, 1997
;;; Revised: Jun     6, 1997
;;;

;;;
;;; Loading.
;;;
(provide 'tui-game)
(require 'tuitate)
(require 'tui-ipc)
(require 'tui-disp)

;;;
;;; System variables.
;;;
(defvar tuitate-game-wait-flag     nil)
(defvar tuitate-game-notation-list nil)
(defvar tuitate-game-candidates    nil)
(defvar tuitate-game-hold          nil)
(defvar tuitate-game-position      nil)
(defvar tuitate-game-input         nil)
(defvar tuitate-game-output        nil)
(defvar tuitate-game-opponent-life nil)
(defvar tuitate-game-your-life     nil)
(defvar tuitate-game-board         nil)
(defvar tuitate-game-belongings    nil)
(defvar tuitate-game-black-flag    nil)
(defvar tuitate-game-turn-flag     nil)
(defvar tuitate-game-turn-number   nil)
(defvar tuitate-game-turn-string   nil)
(defvar tuitate-game-opponent-name nil)
(defvar tuitate-game-opponent-mail nil)
(defvar tuitate-game-message       nil)
(defvar tuitate-game-mode-map      nil)
(if tuitate-game-mode-map nil
  (let ((map (make-keymap)))
    (suppress-keymap map)
    (define-key map "\ "      'tuitate-game-determine)
    (define-key map "\C-m"    'tuitate-game-determine)
    (define-key map "f"       'tuitate-game-forward-square)
    (define-key map "\C-f"    'tuitate-game-forward-square)
    (define-key map [right]   'tuitate-game-forward-square)
    (define-key map "b"       'tuitate-game-backward-square)
    (define-key map "\C-b"    'tuitate-game-backward-square)
    (define-key map [left]    'tuitate-game-backward-square)
    (define-key map "n"       'tuitate-game-next-line)
    (define-key map "\C-n"    'tuitate-game-next-line)
    (define-key map [down]    'tuitate-game-next-line)
    (define-key map "p"       'tuitate-game-previous-line)
    (define-key map "\C-p"    'tuitate-game-previous-line)
    (define-key map [up]      'tuitate-game-previous-line)
    (define-key map "\C-i"    'tuitate-game-forward-candidate)
    (define-key map "\M-\C-i" 'tuitate-game-backward-candidate)
    (define-key map "a"       'tuitate-game-beginning-of-line)
    (define-key map "\C-a"    'tuitate-game-beginning-of-line)
    (define-key map "e"       'tuitate-game-end-of-line)
    (define-key map "\C-e"    'tuitate-game-end-of-line)
    (define-key map "<"       'tuitate-game-beginning-of-board)
    (define-key map "\M-<"    'tuitate-game-beginning-of-board)
    (define-key map ">"       'tuitate-game-end-of-board)
    (define-key map "\M->"    'tuitate-game-end-of-board)
    (define-key map "l"       'tuitate-game-redisplay)
    (define-key map "\C-l"    'tuitate-game-redisplay)
    (define-key map "q"       'tuitate-game-giveup)
    (define-key map "Q"       'tuitate-game-force-quit)
    (define-key map "w"       'tuitate-game-put-belonging)
    (define-key map "g"       'tuitate-game-cancel)
    (define-key map "0"       'tuitate-game-goto-column)
    (define-key map "1"       'tuitate-game-goto-column)
    (define-key map "2"       'tuitate-game-goto-column)
    (define-key map "3"       'tuitate-game-goto-column)
    (define-key map "4"       'tuitate-game-goto-column)
    (define-key map "5"       'tuitate-game-goto-column)
    (define-key map "6"       'tuitate-game-goto-column)
    (define-key map "7"       'tuitate-game-goto-column)
    (define-key map "8"       'tuitate-game-goto-column)
    (define-key map "9"       'tuitate-game-goto-column)
    (define-key map "\C-c0"   'tuitate-game-goto-line)
    (define-key map "\C-c1"   'tuitate-game-goto-line)
    (define-key map "\C-c2"   'tuitate-game-goto-line)
    (define-key map "\C-c3"   'tuitate-game-goto-line)
    (define-key map "\C-c4"   'tuitate-game-goto-line)
    (define-key map "\C-c5"   'tuitate-game-goto-line)
    (define-key map "\C-c6"   'tuitate-game-goto-line)
    (define-key map "\C-c7"   'tuitate-game-goto-line)
    (define-key map "\C-c8"   'tuitate-game-goto-line)
    (define-key map "\C-c9"   'tuitate-game-goto-line)
    (define-key map "."       'tuitate-game-re-message)
    (define-key map "v"       'tuitate-game-disp-information)
    (define-key map "V"       'tuitate-version)
    (define-key map "\C-cc"   'tuitate-game-cheat)
    
    (setq tuitate-game-mode-map map)))

;;;
;;; Constant variables.
;;;
(defconst tuitate-game-initial-life 8)
(defconst tuitate-game-piece-strings
  '((pawn   . "PAWN")
    (lance  . "LANCE")
    (knight . "KNIGHT")
    (silver . "SILVER")
    (gold   . "GOLD")
    (bishop . "BISHOP")
    (rook   . "ROOK")))
(defconst tuitate-game-crypted-passwd "P49d3+.J")

;;;
;;; Macros.
;;;

;; This macro avoids a bug that remains key bindings.
;; Maybe it is Emacs's bug.
(defmacro tuitate-game-def-command (name arglist doc interactive &rest body)
  (list 'defun name arglist doc interactive
	(nconc (list 'if 
		     (list 'not 
			   (list 'eq 
				 (list 'current-local-map)
				 (list 'eval
				       (list 'quote
					     'tuitate-game-mode-map))))
		     nil)
	       body)))

;; User can display last message again.
(defmacro tuitate-game-message (format &rest args)
  (list 'let (list (list 'msg (nconc (list 'format format) args)))
	(list 'setq 'tuitate-game-message 'msg)
	(list 'message 'msg)))

;;;
;;; Mode function.
;;;

;; tuitate-game-mode is special.
(put 'tuitate-game-mode 'mode-class 'special)

(defun tuitate-game-mode (&optional turn opponent-name opponent-mail)
  "Major mode for playing tuitate shogi.
The first line shows turn information.
The second line shows name information.
The third line shows life information.
The others are the board and your belongings.
\\{tuitate-game-mode-map}
"
  (interactive)
  (kill-all-local-variables)
  (switch-to-buffer tuitate-buffer-name)
  (setq major-mode 'tuitate-game-mode
	mode-name "tuitate-game"
	buffer-read-only t
	truncate-lines t
	tuitate-game-hold nil
	tuitate-game-black-flag turn
	tuitate-game-turn-flag turn
	tuitate-game-wait-flag (not turn)
	tuitate-game-opponent-name opponent-name
	tuitate-game-opponent-mail opponent-mail
	tuitate-game-turn-number 1
	tuitate-game-turn-string "1"
	tuitate-game-board (vector (make-vector 9 nil)
				   (make-vector 9 nil)
				   (make-vector 9 nil)
				   (make-vector 9 nil)
				   (make-vector 9 nil)
				   (make-vector 9 nil)
				   (make-vector 9 'pawn)
				   (vector nil 'bishop nil nil nil nil
					   nil 'rook nil)
				   (vector 'lance 'knight 'silver 'gold 'king
					   'gold 'silver 'knight 'lance))
	tuitate-game-belongings (list (cons 'pawn 0)
				      (cons 'lance 0)
				      (cons 'knight 0)
				      (cons 'silver 0)
				      (cons 'gold 0)
				      (cons 'bishop 0)
				      (cons 'rook 0))
	tuitate-game-opponent-life tuitate-game-initial-life
	tuitate-game-your-life tuitate-game-initial-life
	tuitate-game-message nil)
  (if (equal (nth (- (length mode-line-format) 2) mode-line-format)
	     '(-3 . "%p"))
      (setq mode-line-format
	    (let ((mlf (copy-sequence mode-line-format))
		  (l (length mode-line-format)))
	      (setcdr (nthcdr (- l 3) mlf)
		      '("Turn " tuitate-game-turn-string "-%-"))
	      mlf)))
  (use-local-map tuitate-game-mode-map)
  (if (not (eq (process-status tuitate-ipc-process-name) 'open))
      (progn
	(kill-buffer tuitate-buffer-name)
	(error "Not connected yet.")))
  (tuitate-disp-make-faces)
  (tuitate-game-redisplay)
  (or tuitate-inhibit-beep (beep))
  (tuitate-game-message "Game start! %s turn"
			(if tuitate-game-black-flag
			    "Your"
			  "Opponent"))
  (tuitate-ipc-set-filter 'tuitate-game-filter)
  (setq tuitate-ipc-receive-flag t)
  (tuitate-game-filter nil "")		; deal with buffering messages.
  (run-hooks 'tuitate-game-mode-hook))

(defun tuitate-game-belonging-index (piece)
  (let ((cc (assq piece tuitate-game-belongings)))
    (- (length tuitate-game-belongings)
       (length (memq cc tuitate-game-belongings)))))

;;;
;;; Commands.
;;;
(tuitate-game-def-command
 tuitate-game-redisplay ()
 "Re-display the buffer"
 (interactive)
 (let ((old-buffer (current-buffer))
       (old-marker (copy-marker (mark-marker))))
   (unwind-protect
       (progn
	 (set-buffer tuitate-buffer-name)
	 (let ((buffer-read-only nil))
	   (erase-buffer)
;;	   (tuitate-disp-insert-turn-string (not tuitate-game-black-flag))
	   (tuitate-disp-insert-name-string tuitate-game-opponent-name
					    tuitate-player-name
					    t)
	   (tuitate-disp-write-life-string tuitate-game-opponent-life
					   tuitate-game-your-life)
	   (tuitate-disp-insert-board (not tuitate-game-black-flag))
	   (tuitate-disp-write-belongings nil tuitate-game-belongings)
	   (let ((x 0)
		 (y 0)
		 piece)
	     (while (< y 9)
	       (setq x 0)
	       (while (< x 9)
		 (setq piece (aref (aref tuitate-game-board y) x))
		 (tuitate-disp-write-piece piece x y)
		 (if (eq piece 'king)
		     (setq tuitate-game-position (cons x y)))
		 (setq x (1+ x)))
	       (setq y (1+ y))))
	   (tuitate-disp-turn (not tuitate-game-turn-flag)
			      (not tuitate-game-black-flag)
			      nil
			      t)
	   (goto-char (tuitate-disp-piece-point (car tuitate-game-position)
						(cdr tuitate-game-position)))
	   ))
     
     (set-buffer old-buffer)
     (set-marker (mark-marker) old-marker))))

(tuitate-game-def-command
 tuitate-game-forward-square ()
 "Move forward square"
 (interactive)
 (cond ((symbolp tuitate-game-position)
	(let ((pos
	       (cons 0
		     (+ 3 (tuitate-game-belonging-index
			   tuitate-game-position)))))
	  (if (= (cdr pos) 9)
	      (error "End of board"))
	  (setq tuitate-game-position pos)
	  (goto-char (tuitate-disp-piece-point (car pos)
					       (cdr pos)))))
       ((and (= 8 (car tuitate-game-position))
	     (>= (cdr tuitate-game-position) 2))
	(let ((index (- (cdr tuitate-game-position) 2)))
	  (setq tuitate-game-position
		(car (nth index tuitate-game-belongings)))
	  (goto-char (tuitate-disp-belonging-point index))))
       (t
	(setcar tuitate-game-position (1+ (car tuitate-game-position)))
	(if (> (car tuitate-game-position) 8)
	    (setq tuitate-game-position
		  (cons 0 (1+ (cdr tuitate-game-position)))))
	(goto-char (tuitate-disp-piece-point (car tuitate-game-position)
					     (cdr tuitate-game-position))))))

(tuitate-game-def-command
 tuitate-game-backward-square ()
 "Move backward square"
 (interactive)
 (cond ((symbolp tuitate-game-position)
	(let ((pos
	       (cons 8
		     (+ 2 (tuitate-game-belonging-index
			   tuitate-game-position)))))
	  (setq tuitate-game-position pos)
	  (goto-char (tuitate-disp-piece-point (car pos)
					       (cdr pos)))))
       ((and (= 0 (car tuitate-game-position))
	     (>= (cdr tuitate-game-position) 3))
	(let ((index (- (cdr tuitate-game-position) 3)))
	  (setq tuitate-game-position
		(car (nth index tuitate-game-belongings)))
	  (goto-char (tuitate-disp-belonging-point index))))
       (t
	(let ((pos (cons (1- (car tuitate-game-position))
			 (cdr tuitate-game-position))))
	  (if (< (car pos) 0)
	      (setq pos
		    (cons 8 (1- (cdr pos)))))
	  (if (< (cdr pos) 0)
	      (error "Beginning of board"))
	  (setq tuitate-game-position pos)
	  (goto-char (tuitate-disp-piece-point (car pos)
					       (cdr pos)))))))

(tuitate-game-def-command
 tuitate-game-next-line ()
 "Move forward line"
 (interactive)
 (cond ((symbolp tuitate-game-position)
	(let ((index (tuitate-game-belonging-index tuitate-game-position)))
	  (if (= index (1- (length tuitate-game-belongings)))
	      (error "End of board"))
	  (setq tuitate-game-position
		(car (nth (1+ index) tuitate-game-belongings)))
	  (goto-char (tuitate-disp-belonging-point (1+ index)))))
       (t
	(if (= (cdr tuitate-game-position) 8)
	    (error "End of board"))
	(setcdr tuitate-game-position (1+ (cdr tuitate-game-position)))
	(goto-char (tuitate-disp-piece-point (car tuitate-game-position)
					     (cdr tuitate-game-position))))))

(tuitate-game-def-command
 tuitate-game-previous-line ()
 "Move backward line"
 (interactive)
 (cond ((symbolp tuitate-game-position)
	(let ((index (tuitate-game-belonging-index tuitate-game-position)))
	  (if (= index 0)
	      (error "Beginning of board"))
	  (setq tuitate-game-position
		(car (nth (1- index) tuitate-game-belongings)))
	  (goto-char (tuitate-disp-belonging-point (1- index)))))
       (t
	(if (= (cdr tuitate-game-position) 0)
	    (error "Beginning of board"))
	(setcdr tuitate-game-position (1- (cdr tuitate-game-position)))
	(goto-char (tuitate-disp-piece-point (car tuitate-game-position)
					     (cdr tuitate-game-position))))))

(tuitate-game-def-command
 tuitate-game-beginning-of-line ()
 "Move the beginning of current line"
 (interactive)
 (cond ((symbolp tuitate-game-position)
	(let ((index (tuitate-game-belonging-index tuitate-game-position)))
	  (setq tuitate-game-position (cons 0 (+ index 2)))
	  (goto-char (tuitate-disp-piece-point (car tuitate-game-position)
					       (cdr tuitate-game-position)))))
       (t
	(setcar tuitate-game-position 0)
	(goto-char (tuitate-disp-piece-point (car tuitate-game-position)
					     (cdr tuitate-game-position))))))

(tuitate-game-def-command
 tuitate-game-end-of-line ()
 "Move the end of current line"
 (interactive)
 (cond ((symbolp tuitate-game-position)
	(let ((index (tuitate-game-belonging-index tuitate-game-position)))
	  (setq tuitate-game-position (cons 8 (+ index 2)))
	  (goto-char (tuitate-disp-piece-point (car tuitate-game-position)
					       (cdr tuitate-game-position)))))
       (t
	(setcar tuitate-game-position 8)
	(goto-char (tuitate-disp-piece-point (car tuitate-game-position)
					     (cdr tuitate-game-position))))))

(tuitate-game-def-command
 tuitate-game-beginning-of-board ()
 "Move the beginning of the board"
 (interactive)
 (setq tuitate-game-position (cons 0 0))
 (goto-char (tuitate-disp-piece-point 0 0)))

(tuitate-game-def-command
 tuitate-game-end-of-board ()
 "Move the end of the board"
 (interactive)
 (setq tuitate-game-position (cons 8 8))
 (goto-char (tuitate-disp-piece-point 8 8)))

(tuitate-game-def-command
 tuitate-game-giveup ()
 "Give the game up"
 (interactive)
 (if (y-or-n-p "Do you want to give up really? ")
     (progn
       (tuitate-ipc-send-giveup)
       (setq tuitate-game-output 'giveup))))

(tuitate-game-def-command
 tuitate-game-force-quit ()
 "Force to quit"
 (interactive)
 (unwind-protect
     (progn
       (if (eq (process-status tuitate-ipc-process-name) 'open)
	   (progn
	     (tuitate-ipc-send-quit)
	     (setq tuitate-game-output 'quit)))
       (tuitate-ipc-shutdown))
   (kill-buffer tuitate-buffer-name)))

;; Make candidates' list.
(defun tuitate-game-make-candidates (hold)
  (setq tuitate-game-candidates nil)
  (cond ((symbolp hold)
	 (let ((cols (list 0 1 2 3 4 5 6 7 8)))
	   (if (eq hold 'pawn)
	       (let ((x 0) (y 0))
		 (while (< x 9)
		   (setq y 0)
		   (while (< y 9)
		     (if (eq (aref (aref tuitate-game-board y) x) 'pawn)
			 (progn
			   (setq cols (delq x cols))
			   (setq y 9)))
		     (setq y (1+ y)))
		   (setq x (1+ x)))))
	   (mapcar (function
		    (lambda (x)
		      (let ((y 0))
			(cond ((eq hold 'pawn)
			       (setq y 1))
			      ((eq hold 'lance)
			       (setq y 1))
			      ((eq hold 'knight)
			       (setq y 2)))
			(while (< y 9)
			  (if (eq (aref (aref tuitate-game-board y) x) nil)
			      (setq tuitate-game-candidates
				    (append tuitate-game-candidates
					    (cons (cons x y) nil))))
			  (setq y (1+ y))))))
		   cols)))
	(t
	 (let ((piece
		(aref (aref tuitate-game-board
			    (cdr hold))
		      (car hold))))
	   (cond ((eq piece 'pawn)
		  (if (not (aref (aref tuitate-game-board
				       (1- (cdr hold)))
				 (car hold)))
		      (setq tuitate-game-candidates
			    (list (cons (car hold)
					(1- (cdr hold)))))))
		 ((eq piece 'lance)
		  (let ((y (1- (cdr hold))) (x (car hold)))
		    (while (and (>= y 0)
				(null (aref (aref tuitate-game-board y) x)))
		      (setq tuitate-game-candidates
			    (append tuitate-game-candidates
				    (cons (cons x y) nil)))
		      (setq y (1- y)))))
		 ((eq piece 'knight)
		  (let ((y (- (cdr hold) 2)))
		    (mapcar (function
			     (lambda (x)
			       (if (and (>= x 0)
					(< x 9)
					(not (aref
					      (aref tuitate-game-board y)
					      x)))
				   (setq tuitate-game-candidates
					 (append tuitate-game-candidates
						 (cons (cons x y) nil))))))
			    (list (1- (car hold)) (1+ (car hold))))))
		 ((eq piece 'silver)
		  (let ((sy (cdr hold))
			(sx (car hold)))
		    (mapcar (function
			     (lambda (item)
			       (let ((x (car item))
				     (y (cdr item)))
				 (if (and (>= x 0)
					  (< x 9)
					  (>= y 0)
					  (< y 9)
					  (not (aref
						(aref tuitate-game-board y)
						x)))
				     (setq tuitate-game-candidates
					   (append tuitate-game-candidates
						   (cons item nil)))))))
			    (list (cons (1- sx) (1- sy))
				  (cons sx (1- sy))
				  (cons (1+ sx) (1- sy))
				  (cons (1- sx) (1+ sy))
				  (cons (1+ sx) (1+ sy))))))
		 ((eq piece 'king)
		  (let ((sy (cdr hold))
			(sx (car hold)))
		    (mapcar (function
			     (lambda (item)
			       (let ((x (car item))
				     (y (cdr item)))
				 (if (and (>= x 0)
					  (< x 9)
					  (>= y 0)
					  (< y 9)
					  (not (aref
						(aref tuitate-game-board y)
						x)))
				     (setq tuitate-game-candidates
					   (append tuitate-game-candidates
						   (cons item nil)))))))
			    (list (cons (1- sx) (1- sy))
				  (cons sx (1- sy))
				  (cons (1+ sx) (1- sy))
				  (cons (1+ sx) sy)
				  (cons (1- sx) sy)
				  (cons (1- sx) (1+ sy))
				  (cons sx (1+ sy))
				  (cons (1+ sx) (1+ sy))))))
		 ((or (eq piece 'bishop)
		      (eq piece 'Bishop))
		  (let ((sx (car hold))
			(sy (cdr hold))
			x
			y)
		    (mapcar (function
			     (lambda (item)
			       (setq x (+ sx (car item))
				     y (+ sy (cdr item)))
			       (while (and (>= x 0)
					   (< x 9)
					   (>= y 0)
					   (< y 9)
					   (null
					    (aref (aref
						   tuitate-game-board
						   y)
						  x)))
				 (setq tuitate-game-candidates
				       (append tuitate-game-candidates
					       (cons (cons x y) nil)))
				 (setq x (+ x (car item))
				       y (+ y (cdr item))))))
			    '((-1 . -1) (-1 . 1) (1 . -1) (1 . 1)))
		    (if (eq piece 'Bishop)
			(mapcar (function
				 (lambda (item)
				   (setq x (car item)
					 y (cdr item))
				   (if (and (>= x 0)
					    (< x 9)
					    (>= y 0)
					    (< y 9)
					    (null (aref (aref
							 tuitate-game-board
							 y)
							x)))
				       (setq tuitate-game-candidates
					     (append tuitate-game-candidates
						     (cons (cons x y) nil))))))
				(list (cons sx (1- sy))
				      (cons sx (1+ sy))
				      (cons (1- sx) sy)
				      (cons (1+ sx) sy))))))
		 ((or (eq piece 'rook)
		      (eq piece 'Rook))
		  (let ((sx (car hold))
			(sy (cdr hold))
			x
			y)
		    (mapcar (function
			     (lambda (item)
			       (setq x (+ sx (car item))
				     y (+ sy (cdr item)))
			       (while (and (>= x 0)
					   (< x 9)
					   (>= y 0)
					   (< y 9)
					   (null
					    (aref (aref
						   tuitate-game-board
						   y)
						  x)))
				 (setq tuitate-game-candidates
				       (append tuitate-game-candidates
					       (cons (cons x y) nil)))
				 (setq x (+ x (car item))
				       y (+ y (cdr item))))))
			    '((-1 . 0) (1 . 0) (0 . -1) (0 . 1)))
		    (if (eq piece 'Rook)
			(mapcar (function
				 (lambda (item)
				   (setq x (car item)
					 y (cdr item))
				   (if (and (>= x 0)
					    (< x 9)
					    (>= y 0)
					    (< y 9)
					    (null (aref (aref
							 tuitate-game-board
							 y)
							x)))
				       (setq tuitate-game-candidates
					     (append tuitate-game-candidates
						     (cons (cons x y) nil))))))
				(list (cons (1- sx) (1- sy))
				      (cons (1- sx) (1+ sy))
				      (cons (1+ sx) (1- sy))
				      (cons (1+ sx) (1+ sy)))))))
		 (t
		  ;; gold, and promoted pieces (Piece, Lance, Knight, Silver)
		  (let ((sy (cdr hold))
			(sx (car hold)))
		    (mapcar (function
			     (lambda (item)
			       (let ((x (car item))
				     (y (cdr item)))
				 (if (and (>= x 0)
					  (< x 9)
					  (>= y 0)
					  (< y 9)
					  (not (aref
						(aref tuitate-game-board y)
						x)))
				     (setq tuitate-game-candidates
					   (append tuitate-game-candidates
						   (cons item nil)))))))
			    (list (cons (1- sx) (1- sy))
				  (cons sx (1- sy))
				  (cons (1+ sx) (1- sy))
				  (cons (1+ sx) sy)
				  (cons (1- sx) sy)
				  (cons sx (1+ sy))))))))))
  (setq tuitate-game-candidates
	(sort tuitate-game-candidates
	      (function (lambda (p1 p2)
			  (if (= (cdr p1) (cdr p2))
			      (< (car p1) (car p2))
			    (< (cdr p1) (cdr p2))))))))

(tuitate-game-def-command
 tuitate-game-determine ()
 "Hold and put"
 (interactive)
 (if tuitate-game-wait-flag
     (error "Wait just a moment"))
 (catch 'tuitate-game-determine
   (cond ((null tuitate-game-hold)
	  (cond ((symbolp tuitate-game-position)
		 (let ((num (cdr (assq tuitate-game-position
				       tuitate-game-belongings))))
		   (if (= num 0)
		       (error "You don\'t have this piece"))
		   (tuitate-game-make-candidates tuitate-game-position)
		   (if (null tuitate-game-candidates)
		       (error "Where do you drop it?"))
		   (setq tuitate-game-hold tuitate-game-position)
		   (tuitate-disp-hold-belonging
		    (tuitate-game-belonging-index tuitate-game-hold))))
		(t
		 (let ((piece
			(aref (aref tuitate-game-board
				    (cdr tuitate-game-position))
			      (car tuitate-game-position))))
		   (if (null piece)
		       (error "Are you stupid?"))
		   (tuitate-game-make-candidates tuitate-game-position)
		   (if (null tuitate-game-candidates)
		       (error "Where do you move it?"))
		   (setq tuitate-game-hold
			 (cons (car tuitate-game-position)
			       (cdr tuitate-game-position)))
		   (tuitate-disp-hold-piece (car tuitate-game-position)
					    (cdr tuitate-game-position)))))
	  (mapcar (function (lambda (item)
			      (tuitate-disp-candidate (car item)
						      (cdr item))))
		  tuitate-game-candidates))
	 ((symbolp tuitate-game-hold)
	  (if (eq tuitate-game-position tuitate-game-hold)
	      (progn
		(tuitate-game-cancel)
		(throw 'tuitate-game-determine nil)))
	  (if (not (member tuitate-game-position tuitate-game-candidates))
	      (error "Oh, you cannot drop it here"))
	  (let ((dx (car tuitate-game-position))
		(dy (cdr tuitate-game-position)))
	    (tuitate-ipc-send-drop (if tuitate-game-black-flag
				       (- 9 dx)
				     (1+ dx))
				   (if tuitate-game-black-flag
				       (1+ dy)
				     (- 9 dy))
				   (cdr (assq tuitate-game-hold
					      tuitate-game-piece-strings)))
	    (setq tuitate-game-output (list 'drop dx dy tuitate-game-hold)
		  tuitate-game-wait-flag t)))
	 (t
	  (if (equal tuitate-game-position tuitate-game-hold)
	      (progn
		(tuitate-game-cancel)
		(throw 'tuitate-game-determine nil)))
	  (if (not (member tuitate-game-position tuitate-game-candidates))
	      (error "It\'s impossible to move it here!"))
	  (let* ((sx (car tuitate-game-hold))
		 (sy (cdr tuitate-game-hold))
		 (dx (car tuitate-game-position))
		 (dy (cdr tuitate-game-position))
		 (promote nil)
		 (piece (aref (aref tuitate-game-board sy) sx)))
	    (cond ((or (and (eq piece 'pawn) (= dy 0))
		       (and (eq piece 'lance) (= dy 0))
		       (and (eq piece 'knight) (<= dy 1)))
		   (setq promote t))
		  ((and (memq piece '(pawn lance knight silver bishop rook))
			(or (<= dy 2)
			    (<= sy 2)))
		   (setq promote
			 (y-or-n-p "Do you want to promote it? "))))
	    (tuitate-ipc-send-move (if tuitate-game-black-flag
				       (- 9 sx)
				     (1+ sx))
				   (if tuitate-game-black-flag
				       (1+ sy)
				     (- 9 sy))
				   (if tuitate-game-black-flag
				       (- 9 dx)
				     (1+ dx))
				   (if tuitate-game-black-flag
				       (1+ dy)
				     (- 9 dy))
				   promote)
	    (setq tuitate-game-output (list 'move sx sy dx dy promote)
		  tuitate-game-wait-flag t))))))

(tuitate-game-def-command
 tuitate-game-put-belonging ()
 "Put a belonging. Ask a piece name"
 (interactive)
 (if tuitate-game-wait-flag
     (error "Don\'t hurry!"))
 (let (piece-string piece)
   (setq piece-string
	 (completing-read "Piece name: "
			  '(("pawn" 1) ("lance" 2) ("knight" 3) ("silver" 4)
			    ("gold" 5) ("bishop" 6) ("rook" 7))
			  nil
			  t))
   (if piece-string
       (progn
	 (setq piece (cdr (assoc piece-string
				 '(("pawn"   . pawn)
				   ("lance"  . lance)
				   ("knight" . knight)
				   ("silver" . silver)
				   ("gold"   . gold)
				   ("bishop" . bishop)
				   ("rook"   . rook)))))
	 (if (= (cdr (assq piece tuitate-game-belongings)) 0)
	     (error "You don\'t have this piece"))
	 (tuitate-game-make-candidates piece)
	 (if (not (member tuitate-game-position tuitate-game-candidates))
	     (error "Oh, you cannot drop it here"))
	 (let ((dx (car tuitate-game-position))
	       (dy (cdr tuitate-game-position)))
	   (tuitate-ipc-send-drop (if tuitate-game-black-flag
				      (- 9 dx)
				    (1+ dx))
				  (if tuitate-game-black-flag
				      (1+ dy)
				    (- 9 dy))
				  (cdr (assq piece
					     tuitate-game-piece-strings)))
	   (setq tuitate-game-output (list 'drop dx dy piece)
		 tuitate-game-wait-flag t))))))

;; Erase any overlay.
(defun tuitate-game-erase-overlay ()
  (if tuitate-game-hold
      (progn
	(mapcar (function (lambda (item)
			    (tuitate-disp-candidate (car item) (cdr item) t)))
		tuitate-game-candidates)
	(cond ((symbolp tuitate-game-hold)
	       (tuitate-disp-hold-belonging
		(tuitate-game-belonging-index tuitate-game-hold)
		t))
	      (t
	       (tuitate-disp-hold-piece (car tuitate-game-hold)
					(cdr tuitate-game-hold)
					t))))))

(tuitate-game-def-command
 tuitate-game-cancel ()
 "Cancel to put holding piece"
 (interactive)
 (tuitate-game-erase-overlay)
 (tuitate-game-message "Cancel")
 (setq tuitate-game-hold nil))

(defun tuitate-game-forward-coords (px py)
  (if (< (eval px) 8)
      (set px (1+ (eval px)))
    (set px 0)
    (if (< (eval py) 8)
	(set py (1+ (eval py)))
      (set py 0))))

(tuitate-game-def-command
 tuitate-game-forward-candidate ()
 "Move forward candidate. Only on the board"
 (interactive)
 (let ((x (if (symbolp tuitate-game-position)
	      8
	    (car tuitate-game-position)))
       (y (if (symbolp tuitate-game-position)
	      (+ 2 (tuitate-game-belonging-index tuitate-game-position))
	    (cdr tuitate-game-position))))
   (tuitate-game-forward-coords 'x 'y)
   (cond ((null tuitate-game-hold)
	  (while (null (aref (aref tuitate-game-board y) x))
	    (tuitate-game-forward-coords 'x 'y))
	  (setq tuitate-game-position (cons x y))
	  (goto-char (tuitate-disp-piece-point x y)))
	 (t
	  (setq tuitate-game-position
		(catch 'find-candidate
		  (mapcar (function (lambda (item)
				      (if (or (and (<= x (car item))
						   (= y (cdr item)))
					      (< y (cdr item)))
					  (throw 'find-candidate
						 (cons (car item)
						       (cdr item))))))
			  tuitate-game-candidates)
		  (cons (car (car tuitate-game-candidates))
			(cdr (car tuitate-game-candidates)))))
	  (goto-char (tuitate-disp-piece-point
		      (car tuitate-game-position)
		      (cdr tuitate-game-position)))))))

(defun tuitate-game-backward-coords (px py)
  (if (> (eval px) 0)
      (set px (1- (eval px)))
    (set px 8)
    (if (> (eval py) 0)
	(set py (1- (eval py)))
      (set py 8))))

(tuitate-game-def-command
 tuitate-game-backward-candidate ()
 "Move backward candidates. Only on the board"
 (interactive)
 (let ((x (if (symbolp tuitate-game-position)
	      8
	    (car tuitate-game-position)))
       (y (if (symbolp tuitate-game-position)
	      (+ 2 (tuitate-game-belonging-index tuitate-game-position))
	    (cdr tuitate-game-position))))
   (tuitate-game-backward-coords 'x 'y)
   (cond ((null tuitate-game-hold)
	  (while (null (aref (aref tuitate-game-board y) x))
	    (tuitate-game-backward-coords 'x 'y))
	  (setq tuitate-game-position (cons x y))
	  (goto-char (tuitate-disp-piece-point x y)))
	 (t
	  (setq tuitate-game-position
		(catch 'find-candidate
		  (mapcar (function (lambda (item)
				      (if (or (and (>= x (car item))
						   (= y (cdr item)))
					      (> y (cdr item)))
					  (throw 'find-candidate
						 (cons (car item)
						       (cdr item))))))
			  tuitate-game-candidates)
		  (cons (car (nth (1- (length tuitate-game-candidates))
				  tuitate-game-candidates))
			(cdr (nth (1- (length tuitate-game-candidates))
				  tuitate-game-candidates)))))
	  (goto-char (tuitate-disp-piece-point
		      (car tuitate-game-position)
		      (cdr tuitate-game-position)))))))

(tuitate-game-def-command
 tuitate-game-goto-column ()
 (interactive)
 (let ((digit (string-to-number (char-to-string last-command-event))))
   (cond ((zerop digit)
	  ;; Move to your king position.
	  (catch 'find-king
	    (let ((x 0) (y 0))
	      (while (< y 9)
		(setq x 0)
		(while (< x 9)
		  (if (eq (aref (aref tuitate-game-board y) x) 'king)
		      (progn
			(setq tuitate-game-position (cons x y))
			(goto-char (tuitate-disp-piece-point x y))
			(throw 'find-king nil)))
		  (setq x (1+ x)))
		(setq y (1+ y))))))
	 (t
	  (setq digit (if tuitate-game-black-flag
			  (- 9 digit)
			(1- digit)))
	  (if (symbolp tuitate-game-position)
	      (setq tuitate-game-position
		    (cons digit
			  (+ (tuitate-game-belonging-index
			      tuitate-game-position)
			     2)))
	    (setq tuitate-game-position (cons digit
					      (cdr tuitate-game-position))))
	  (goto-char (tuitate-disp-piece-point digit
					       (cdr tuitate-game-position)))
	  ))))
 
(tuitate-game-def-command
 tuitate-game-goto-line ()
 (interactive)
 (let ((digit (string-to-number (char-to-string last-command-event))))
   (cond ((zerop digit)
	  ;; Move to your king position.
	  (catch 'find-king
	    (let ((x 0) (y 0))
	      (while (< y 9)
		(setq x 0)
		(while (< x 9)
		  (if (eq (aref (aref tuitate-game-board y) x) 'king)
		      (progn
			(setq tuitate-game-position (cons x y))
			(goto-char (tuitate-disp-piece-point x y))
			(throw 'find-king nil)))
		  (setq x (1+ x)))
		(setq y (1+ y))))))
	 (t
	  (setq digit (if tuitate-game-black-flag
			  (1- digit)
			(- 9 digit)))
	  (if (symbolp tuitate-game-position)
	      (setq tuitate-game-position
		    (cons 8 digit))
	    (setq tuitate-game-position (cons (car tuitate-game-position)
					      digit)))
	  (goto-char (tuitate-disp-piece-point (car tuitate-game-position)
					       digit))
	  ))))

(tuitate-game-def-command
 tuitate-game-re-message ()
 "Display last message again"
 (interactive)
 (if tuitate-game-message
     (message tuitate-game-message)))

(tuitate-game-def-command
 tuitate-game-disp-information ()
 "Display your opponent's information"
 (interactive)
 (message "%s: %s <%s>"
	  tuitate-opponent-string
	  tuitate-game-opponent-name
	  tuitate-game-opponent-mail))

;; Crypt a password.
(defun tuitate-game-crypt (passwd)
  (let ((len (length passwd))
	(secret "")
	(seed 4274)
	(index 0))
    (while (< index len)
      (setq seed (+ (* seed (aref passwd index) 7401847))
	    index (1+ index)))
    (setq index 0)
    (while (< index 8)
      (if (< index len)
	  (setq seed (logxor (aref passwd index) seed))
	(setq seed (+ seed (aref passwd (mod index len)))))
      (setq seed (+ (* seed 474710701) 1101079))
      (let ((ch (mod seed 128)))
	(while (not (string-match "^[a-zA-Z0-9\\-\\+\\.,!&]$" 
				  (char-to-string ch)))
	  (setq seed (+ (* seed 118235) 74110823))
	  (setq ch (mod seed 128)))
	(setq secret (concat secret (char-to-string ch))))
      (setq index (1+ index)))
    secret))

(tuitate-game-def-command
 tuitate-game-cheat (passwd)
 "Display cheating information at current point"
 (interactive "sPassword: ")
 (if (symbolp tuitate-game-position)
     (error "Must be on board"))
 (let ((num (- 50 (length passwd))))
   (if (< num 1)
       (error "Too long password"))
   (while (>= num 0)
     (setq passwd (tuitate-game-crypt passwd)
	   num (1- num)))
   (if (not (string= passwd tuitate-game-crypted-passwd))
       (error "Incorrect password")))
 (let ((dx (car tuitate-game-position))
       (dy (cdr tuitate-game-position)))
   (setq dx (if tuitate-game-black-flag
		(- 9 dx)
	      (1+ dx))
	 dy (if tuitate-game-black-flag
		(1+ dy)
	      (- 9 dy)))
   (tuitate-ipc-send-cheat dx dy)
   (setq tuitate-game-output (list 'cheat dx dy)
	 tuitate-game-wait-flag t)))

;;;
;;; Mode filter.
;;;
(defun tuitate-game-filter (proc string)
  (let ((msg (tuitate-ipc-receive string)))
    (while msg
      (let ((name (car msg)))
	(cond ((and (consp tuitate-game-input)
		    (eq (car tuitate-game-input) 'end))
	       (if (string= name ".")
		   (let ((result (nth 1 tuitate-game-input))
			 (reason (nth 2 tuitate-game-input)))
		     (setq tuitate-game-input nil)
		     (tuitate-ipc-shutdown)
		     (or tuitate-inhibit-beep (beep))
		     (if (y-or-n-p
			  (concat (cdr (assq result
					     '((win  . "You win. ")
					       (lose . "Opponent win. ")
					       (draw . "Draw game. "))))
				  (cdr (assq reason
					     '((checkmate . "Checkmate. ")
					       (loselife  . "Lose life. ")
					       (giveup    . "Give up. ")
					       (emergency . "Emergency. "))))
				  "Do you want to see this notation? "))
			 (progn
			   (require 'tui-not)
			   (tuitate-notation-mode
			    tuitate-game-opponent-name
			    tuitate-game-opponent-mail
			    tuitate-player-name
			    tuitate-player-mail
			    tuitate-game-notation-list
			    (not tuitate-game-black-flag)))
		       (kill-buffer tuitate-buffer-name)))
		 (setq tuitate-game-notation-list
		       (append tuitate-game-notation-list (cons msg nil)))))
	      
	      ((string= name "OK")
	       (let ((notify "Ok"))
		 (setq tuitate-game-input 'ok
		       tuitate-game-turn-flag nil
		       tuitate-game-turn-number (1+ tuitate-game-turn-number)
		       tuitate-game-turn-string (number-to-string
						 tuitate-game-turn-number))
		 (tuitate-game-erase-overlay)
		 (setq tuitate-game-hold nil)
		 (cond ((eq (car tuitate-game-output) 'move)
			(let* ((sx (nth 1 tuitate-game-output))
			       (sy (nth 2 tuitate-game-output))
			       (dx (nth 3 tuitate-game-output))
			       (dy (nth 4 tuitate-game-output))
			       (promote (nth 5 tuitate-game-output))
			       (piece (aref (aref tuitate-game-board sy) sx)))
			  (if promote
			      (setq piece
				    (cdr (assq piece
					       '((pawn   . Pawn)
						 (lance  . Lance)
						 (knight . Knight)
						 (silver . Silver)
						 (bishop . Bishop)
						 (rook   . Rook))))))
			  (aset (aref tuitate-game-board sy) sx nil)
			  (aset (aref tuitate-game-board dy) dx piece)
			  (tuitate-disp-write-piece nil sx sy)
			  (tuitate-disp-write-piece piece dx dy)))
		       ((eq (car tuitate-game-output) 'drop)
			(let* ((dx (nth 1 tuitate-game-output))
			       (dy (nth 2 tuitate-game-output))
			       (piece (nth 3 tuitate-game-output))
			       (belonging (assq piece
						tuitate-game-belongings)))
			  (setcdr belonging (1- (cdr belonging)))
			  (aset (aref tuitate-game-board dy) dx piece)
			  (tuitate-disp-write-belongings
			   nil
			   tuitate-game-belongings)
			  (tuitate-disp-write-piece piece dx dy)))
		       (t
			(error "invalid message")))
		 (if (and (> (length msg) 1)
			  (not (string= (nth 1 msg) "CHECK")))
		     (progn
		       (setq notify (concat notify
					    ". You get a "
					    (nth 1 msg)))
		       (let ((belonging (assq
					 (car (rassoc
					       (nth 1 msg)
					       tuitate-game-piece-strings))
					 tuitate-game-belongings)))
			 (setcdr belonging (1+ (cdr belonging)))
			 (tuitate-disp-write-belongings
			  nil
			  tuitate-game-belongings))))
		 (if (string= (nth (1- (length msg)) msg) "CHECK")
		     (setq notify (concat notify
					  ". You check opponent's king")))
		 (tuitate-game-message notify)
		 (tuitate-disp-turn nil (not tuitate-game-black-flag) t t)
		 (tuitate-disp-turn t (not tuitate-game-black-flag) nil t))
	       (let ((old-buffer (current-buffer))
		     (old-marker (copy-marker (mark-marker))))
		 (unwind-protect
		     (progn
		       (set-buffer tuitate-buffer-name)
		       (goto-char (if (symbolp tuitate-game-position)
				      (tuitate-disp-belonging-point
				       (tuitate-game-belonging-index
					tuitate-game-position))
				    (tuitate-disp-piece-point
				     (car tuitate-game-position)
				     (cdr tuitate-game-position)))))
		   (set-buffer old-buffer)
		   (set-marker (mark-marker) old-marker))))
	      
	      ((string= name "ILLEGAL")
	       (setq tuitate-game-input 'illegal
		     tuitate-game-wait-flag nil)
	       (or tuitate-inhibit-beep (beep))
	       (tuitate-game-erase-overlay)
	       (setq tuitate-game-hold nil
		     tuitate-game-your-life (1- tuitate-game-your-life))
	       (tuitate-disp-write-life-string tuitate-game-opponent-life
					       tuitate-game-your-life)
	       (tuitate-game-message
		"You mistake. But this is the program's bug, too"))
	      
	      ((string= name "VIOLATE")
	       (setq tuitate-game-input 'violate
		     tuitate-game-wait-flag nil)
	       (or tuitate-inhibit-beep (beep))
	       (tuitate-game-erase-overlay)
	       (setq tuitate-game-hold nil
		     tuitate-game-your-life (1- tuitate-game-your-life))
	       (tuitate-disp-write-life-string tuitate-game-opponent-life
					       tuitate-game-your-life)
	       (tuitate-game-message (concat
			 "You mistake. "
			 (cdr (assoc
			       (nth 1 msg)
			       '(("JUMPOVER" . "jump opponent's piece over")
				 ("CHECK"    . "your king is checked")
				 ("DROPON"   . "drop on opponent's piece")
				 ("CHECKMATEBYDROPPINGPAWN" .
				  "checkmate by dropping a pawn")))))))
	      
	      ((string= name "NEXT")
	       (setq tuitate-game-input 'next
		     tuitate-game-turn-flag t
		     tuitate-game-wait-flag nil
		     tuitate-game-turn-number (1+ tuitate-game-turn-number)
		     tuitate-game-turn-string (number-to-string
					       tuitate-game-turn-number))
	       (let ((notify "Your turn"))
		 (if (and (> (length msg) 1)
			  (not (string= (nth 1 msg) "CHECKED")))
		     (let ((x (string-to-number (nth 1 msg)))
			   (y (string-to-number (nth 2 msg)))
			   piece)
		       (setq x (if tuitate-game-black-flag
				   (- 9 x)
				 (1- x))
			     y (if tuitate-game-black-flag
				   (1- y)
				 (- 9 y))
			     piece (aref (aref tuitate-game-board y) x))
		       (aset (aref tuitate-game-board y) x nil)
		       (tuitate-disp-write-piece nil x y)
		       (setq notify (concat notify
					    ". Opponent gets "
					    (nth 1 msg)
					    (cdr (assoc (nth 2 msg)
							'(("1" . "a")
							  ("2" . "b")
							  ("3" . "c")
							  ("4" . "d")
							  ("5" . "e")
							  ("6" . "f")
							  ("7" . "g")
							  ("8" . "h")
							  ("9" . "i"))))
					    " "
					 (cdr (assq
					       piece
					       '((pawn   . "pawn")
						 (lance  . "lance")
						 (knight . "knight")
						 (silver . "silver")
						 (gold   . "gold")
						 (king   . "king")
						 (bishop . "bishop")
						 (rook   . "rook")
						 (Pawn   . "promoting pawn")
						 (Lance  . "promoting lance")
						 (Knight . "promoting knight")
						 (Silver . "promoting silver")
						 (Bishop . "promoting bishop")
						 (Rook   . "promoting rook"))))
					 ))))
		 
		 (if (string= (nth (1- (length msg)) msg) "CHECKED")
		     (setq notify (concat notify
					  ". You are checked")))
		 (or tuitate-inhibit-beep (beep))
		 (tuitate-game-message notify)
		 (tuitate-disp-turn nil (not tuitate-game-black-flag) nil t)
		 (tuitate-disp-turn t (not tuitate-game-black-flag) t t))
	       (let ((old-buffer (current-buffer))
		     (old-marker (copy-marker (mark-marker))))
		 (unwind-protect
		     (progn
		       (set-buffer tuitate-buffer-name)
		       (goto-char (if (symbolp tuitate-game-position)
				      (tuitate-disp-belonging-point
				       (tuitate-game-belonging-index
					tuitate-game-position))
				    (tuitate-disp-piece-point
				     (car tuitate-game-position)
				     (cdr tuitate-game-position)))))
		   (set-buffer old-buffer)
		   (set-marker (mark-marker) old-marker))))
	      
	      ((string= name "ENEMYVIOLATE")
	       (setq tuitate-game-input 'enemyviolate
		     tuitate-game-opponent-life
		     (1- tuitate-game-opponent-life))
	       (tuitate-disp-write-life-string tuitate-game-opponent-life
					       tuitate-game-your-life)
	       (or tuitate-inhibit-beep (beep))
	       (tuitate-game-message "Opponent violates"))
	      
	      ((string= name "END")
	       (let ((result (cdr (assoc (nth 1 msg)
					 '(("WIN"  . win)
					   ("LOSE" . lose)
					   ("DRAW" . draw)))))
		     (reason (cdr (assoc (nth 2 msg)
					 '(("CHECKMATE" . checkmate)
					   ("LOSELIFE"  . loselife)
					   ("GIVEUP"    . giveup)
					   ("EMERGENCY" . emergency))))))
		 (setq tuitate-game-input (list 'end result reason)
		       tuitate-game-wait-flag t
		       tuitate-game-notation-list nil)))
	      
	      ((string= name "CHEAT")
	       (setq tuitate-game-input 'cheat
		     tuitate-game-wait-flag nil)
	       (let ((notify (format "%d%c "
				     (nth 1 tuitate-game-output)
				     (cdr (assq (nth 2 tuitate-game-output)
						'((1 . ?a)
						  (2 . ?b)
						  (3 . ?c)
						  (4 . ?d)
						  (5 . ?e)
						  (6 . ?f)
						  (7 . ?g)
						  (8 . ?h)
						  (9 . ?i)))))))
		 (cond ((string= (nth 1 msg) "BLACK")
			(setq notify (concat notify
					     (if tuitate-game-black-flag
						 "your "
					       "opponent's "))))
		       ((string= (nth 1 msg) "WHITE")
			(setq notify (concat notify
					     (if tuitate-game-black-flag
						 "opponent's "
					       "your "))))
		       ((string= (nth 1 msg) "NONE")
			(setq notify (concat notify
					     "none")))
		       (t
			(error "Invalid message")))
		 (if (not (string= (nth 1 msg) "NONE"))
		     (progn
		       (if (string= (nth 3 msg) "PROMOTE")
			   (setq notify (concat notify "promoting ")))
		       (setq notify (concat notify (downcase (nth 2 msg))))
		       ))
		 (tuitate-game-message notify)))
	      
	      (t
	       (error "Unknown message"))))
      (setq msg (tuitate-ipc-receive "")))))

;;;
;;; End of tui-game.el
;;;
