;;;			     tui-not.el
;;;
;;;	      Copyright (C) 1997       OKUJI Yoshinori
;;;
;;;		   This emacs lisp library conforms
;;;		GNU GENERAL PUBLIC LICENSE Version 2.
;;;
;;; Author:  OKUJI Yoshinori <okuji@kmc.kyoto-u.ac.jp>
;;; Created: May    15, 1997
;;; Revised: May    26, 1997
;;;

;;;
;;; Loading.
;;;
(provide 'tui-not)
(require 'tuitate)
(require 'tui-disp)

;;;
;;; System variables.
;;;
(defvar tuitate-notation-current-data  0)
(defvar tuitate-notation-last-data     0)
(defvar tuitate-notation-turn-string   nil)
(defvar tuitate-notation-list          nil)
(defvar tuitate-notation-reverse       nil)
(defvar tuitate-notation-left-name     nil)
(defvar tuitate-notation-left-mail     nil)
(defvar tuitate-notation-right-name    nil)
(defvar tuitate-notation-right-mail    nil)
(defvar tuitate-notation-compiled-data nil)
(defvar tuitate-notation-mode-map      nil)
(if tuitate-notation-mode-map nil
  (let ((map (make-keymap)))
    (suppress-keymap map)
    (define-key map "f"     'tuitate-notation-forward-turn)
    (define-key map "\C-f"  'tuitate-notation-forward-turn)
    (define-key map [right] 'tuitate-notation-forward-turn)
    (define-key map "\ "    'tuitate-notation-forward-turn)
    (define-key map "b"     'tuitate-notation-backward-turn)
    (define-key map "\C-b"  'tuitate-notation-backward-turn)
    (define-key map [left]  'tuitate-notation-backward-turn)
    (define-key map "n"     'tuitate-notation-next-turn)
    (define-key map "\C-n"  'tuitate-notation-next-turn)
    (define-key map [down]  'tuitate-notation-next-turn)
    (define-key map "\C-m"  'tuitate-notation-next-turn)
    (define-key map "p"     'tuitate-notation-previous-turn)
    (define-key map "\C-p"  'tuitate-notation-previous-turn)
    (define-key map [up]    'tuitate-notation-previous-turn)
    (define-key map "<"     'tuitate-notation-beginning-of-turn)
    (define-key map "\M-<"  'tuitate-notation-beginning-of-turn)
    (define-key map ">"     'tuitate-notation-end-of-turn)
    (define-key map "\M->"  'tuitate-notation-end-of-turn)
    (define-key map "l"     'tuitate-notation-redisplay)
    (define-key map "\C-l"  'tuitate-notation-redisplay)
    (define-key map "q"     'tuitate-notation-quit)
    (define-key map "g"     'tuitate-notation-goto-turn)
    (define-key map "."     'tuitate-notation-re-message)
    (define-key map "v"     'tuitate-notation-disp-information)
    (define-key map "V"     'tuitate-version)
    
    (setq tuitate-notation-mode-map map)))

;;;
;;; Constant variables.
;;;

;; notation mode compiles notation list.
;; This compiled data is list, indexed below.
;; Each piece name is one character symbol, or followed by `*'(promoting).
(defconst tuitate-notation-data-turn              0) ; turn number
(defconst tuitate-notation-data-violation         1) ; what violation
(defconst tuitate-notation-data-check             2) ; check flag
(defconst tuitate-notation-data-get               3) ; which piece is gotten
(defconst tuitate-notation-data-promote           4) ; promote flag
(defconst tuitate-notation-data-play              5) ; play information
(defconst tuitate-notation-data-left-life         6) ; left life
(defconst tuitate-notation-data-right-life        7) ; right life
(defconst tuitate-notation-data-board             8) ; board status
(defconst tuitate-notation-data-left-belongings   9) ; left belongings
(defconst tuitate-notation-data-right-belongings 10) ; right belongings
(defconst tuitate-notation-data-nth              11) ; element of source list
(defconst tuitate-notation-data-message          12) ; message string

;;;
;;; Macros.
;;;

;; I don't use cl.
(defmacro tuitate-notation-setcar (list index obj)
  (list 'setcar (list 'nthcdr index list) obj))

;; This macro avoids a bug that remains key bindings.
;; Maybe it is Emacs's bug.
(defmacro tuitate-notation-def-command (name arglist doc interactive
					     &rest body)
  (list 'defun name arglist doc interactive
	(nconc (list 'if 
		     (list 'not 
			   (list 'eq 
				 (list 'current-local-map)
				 (list 'eval
				       (list 'quote
					     'tuitate-notation-mode-map))))
		     nil)
	       body)))

;;;
;;; Mode function.
;;;

;; tuitate-notation-mode is special.
(put 'tuitate-notation-mode 'mode-class 'special)

(defun tuitate-notation-mode (&optional left-name left-mail
					right-name right-mail
					notation-list reverse)
  "Major mode for viewing tuitate shogi notation.
The first line shows turn information.
The second line shows name information.
The third line shows life information.
The others are the board and the belongings.
\\{tuitate-notation-mode-map}
"
  (interactive)
  (kill-all-local-variables)
  (switch-to-buffer tuitate-buffer-name)
  (setq major-mode 'tuitate-notation-mode
	mode-name "tuitate-notation"
	buffer-read-only t
	truncate-lines t)
  (if (equal (nth (- (length mode-line-format) 2) mode-line-format)
	     '(-3 . "%p"))
      (setq mode-line-format
	    (let ((mlf (copy-sequence mode-line-format))
		  (l (length mode-line-format)))
	      (setcdr (nthcdr (- l 3) mlf)
		      '("Turn " tuitate-notation-turn-string "-%-"))
	      mlf)))
  (use-local-map tuitate-notation-mode-map)
  (setq tuitate-notation-turn-string "*"
	tuitate-notation-list notation-list
	tuitate-notation-reverse reverse
	tuitate-notation-left-name left-name
	tuitate-notation-left-mail left-mail
	tuitate-notation-right-name right-name
	tuitate-notation-right-mail right-mail)
  (tuitate-notation-init-compile)
  (tuitate-disp-make-faces)
  (tuitate-notation-redisplay)
  (run-hooks 'tuitate-notation-mode-hook))

;; Initialize compile status.
(defun tuitate-notation-init-compile ()
  (let (new)
    (setq tuitate-notation-compiled-data (make-vector
					  (1+ (length tuitate-notation-list))
					  nil)
	  tuitate-notation-current-data 0
	  tuitate-notation-last-data 0
	  new (list
	       0
	       nil
	       nil
	       nil
	       nil
	       nil
	       8
	       8
	       (list 'L  'N  'S  'G  'K  'G  'S  'N  'L
		     nil 'R  nil nil nil nil nil 'B  nil
		     'P  'P  'P  'P  'P  'P  'P  'P  'P
		     nil nil nil nil nil nil nil nil nil
		     nil nil nil nil nil nil nil nil nil
		     nil nil nil nil nil nil nil nil nil
		     'p  'p  'p  'p  'p  'p  'p  'p  'p
		     nil 'b  nil nil nil nil nil 'r  nil
		     'l  'n  's  'g  'k  'g  's  'n  'l )
	       (list (cons 'pawn 0) (cons 'lance 0) (cons 'knight 0)
		     (cons 'silver 0) (cons 'gold 0) (cons 'bishop 0)
		     (cons 'rook 0))
	       (list (cons 'pawn 0) (cons 'lance 0) (cons 'knight 0)
		     (cons 'silver 0) (cons 'gold 0) (cons 'bishop 0)
		     (cons 'rook 0))
	       -1			; -1 is not invalid, this is helpful.
	       "Game start!"))
    (aset tuitate-notation-compiled-data 0 new)
    ))

;; Compile the notation list, as needed.
(defun tuitate-notation-compile ()
  (let ((last (aref tuitate-notation-compiled-data tuitate-notation-last-data))
	new elm elm-index turn left)
    (if (eq (car-safe (nth tuitate-notation-data-play last)) 'end)
	(error "There is no more turn"))
    (setq elm-index (1+ (nth tuitate-notation-data-nth last))
	  elm (nth elm-index tuitate-notation-list)
	  new (list
	       nil
	       nil
	       nil
	       nil
	       nil
	       nil
	       (nth tuitate-notation-data-left-life last)
	       (nth tuitate-notation-data-right-life last)
	       (copy-sequence (nth tuitate-notation-data-board last))
	       (copy-alist (nth tuitate-notation-data-left-belongings last))
	       (copy-alist (nth tuitate-notation-data-right-belongings last))
	       elm-index
	       nil)
	  turn (nth tuitate-notation-data-turn last))
    (if (not (nth tuitate-notation-data-violation last))
	(setq turn (1+ turn)))
    (setq left (or (and tuitate-notation-reverse
			(= (logand turn 1) 1))
		   (and (not tuitate-notation-reverse)
			(= (logand turn 1) 0))))
    (cond ((string= (car elm) "MOVE")
	   (let (sx sy dx dy promote check piece violation)
	     (if tuitate-notation-reverse
		 (setq sx (1- (string-to-number (nth 1 elm)))
		       sy (- 9 (string-to-number (nth 2 elm)))
		       dx (1- (string-to-number (nth 3 elm)))
		       dy (- 9 (string-to-number (nth 4 elm))))
	       (setq sx (- 9 (string-to-number (nth 1 elm)))
		     sy (1- (string-to-number (nth 2 elm)))
		     dx (- 9 (string-to-number (nth 3 elm)))
		     dy (1- (string-to-number (nth 4 elm)))))
	     (setq promote (string= (nth 5 elm) "PROMOTE")
		   check (string= (nth 6 elm) "CHECKED"))
	     (if (> (length elm) 7)
		 (setq piece (cdr (assoc (nth 7 elm)
					 '(("PAWN"   . pawn)
					   ("LANCE"  . lance)
					   ("KNIGHT" . knight)
					   ("SILVER" . silver)
					   ("GOLD"   . gold)
					   ("BISHOP" . bishop)
					   ("ROOK"   . rook))))))
	     (if (> (length elm) 7)
		 (setq violation (cdr (assoc (nth (1- (length elm)) elm)
					     '(("JUMPOVER" . jumpover)
					       ("CHECK"    . check)
					       ("DROPON"   . dropon)
					       ("CHECKMATEBYDROPPINGPAWN" .
						pawn))))))
	     (tuitate-notation-setcar new
				      tuitate-notation-data-violation
				      violation)
	     (tuitate-notation-setcar new
				      tuitate-notation-data-check
				      check)
	     (tuitate-notation-setcar new
				      tuitate-notation-data-get
				      piece)
	     (tuitate-notation-setcar new
				      tuitate-notation-data-promote
				      promote)
	     (tuitate-notation-setcar new
				      tuitate-notation-data-play
				      (list 'move sx sy dx dy))
	     (if violation
		 (tuitate-notation-setcar
		  new
		  (if left
		      tuitate-notation-data-left-life
		    tuitate-notation-data-right-life)
		  (1- (nth
		       (if left
			   tuitate-notation-data-left-life
			 tuitate-notation-data-right-life)
		       new))))
	     (let ((board (nth tuitate-notation-data-board new))
		   (left-belongings (nth
				     tuitate-notation-data-left-belongings
				     new))
		   (right-belongings (nth
				      tuitate-notation-data-right-belongings
				      new)))
	       (if piece
		   (if left
		       (setcdr (assq piece left-belongings)
			       (1+ (cdr (assq piece left-belongings))))
		     (setcdr (assq piece right-belongings)
			     (1+ (cdr (assq piece right-belongings))))))
	       (tuitate-notation-setcar new
					tuitate-notation-data-message
					(concat
					 (nth 3 elm)
					 (cdr (assoc (nth 4 elm)
						     '(("1" . "a")
						       ("2" . "b")
						       ("3" . "c")
						       ("4" . "d")
						       ("5" . "e")
						       ("6" . "f")
						       ("7" . "g")
						       ("8" . "h")
						       ("9" . "i"))))
					 " "
					 (cdr (assq
					       (nth
						(+ sx (* sy 9))
						board)
					       '((p . "pawn")
						 (l . "lance")
						 (n . "knight")
						 (s . "silver")
						 (g . "gold")
						 (k . "king")
						 (b . "bishop")
						 (r . "rook")
						 (P . "pawn")
						 (L . "lance")
						 (N . "knight")
						 (S . "silver")
						 (G . "gold")
						 (K . "king")
						 (B . "bishop")
						 (R . "rook")
						 (p* . "promoting pawn")
						 (l* . "promoting lance")
						 (n* . "promoting knight")
						 (s* . "promoting silver")
						 (b* . "promoting bishop")
						 (r* . "promoting rook")
						 (P* . "promoting pawn")
						 (L* . "promoting lance")
						 (N* . "promoting knight")
						 (S* . "promoting silver")
						 (B* . "promoting bishop")
						 (R* . "promoting rook"))))
					 (and promote " promote")
					 ". "
					 (if (and (not violation) piece)
					     (cdr (assq
						   piece
						   '((pawn . "Get pawn. ")
						     (lance . "Get lance. ")
						     (knight . "Get knight. ")
						     (silver . "Get silver. ")
						     (gold . "Get gold. ")
						     (bishop . "Get bishop. ")
						     (rook . "Get rook. ")))))
					 (if violation
					     (cdr
					      (assq
					       violation
					       '((jumpover . "Jumping over. ")
						 (check . "Checked. "))))
					   (if check
					       "Check. "))))
	       (if (not violation)
		   (progn
		     (tuitate-notation-setcar
		      board
		      (+ dx (* dy 9))
		      (if promote
			  (cdr (assq (nth (+ sx (* sy 9))
					  board)
				     '((p . p*)
				       (l . l*)
				       (n . n*)
				       (s . s*)
				       (b . b*)
				       (r . r*)
				       (P . P*)
				       (L . L*)
				       (N . N*)
				       (S . S*)
				       (B . B*)
				       (R . R*))))
			(nth (+ sx (* sy 9)) board)))
		     (tuitate-notation-setcar board
					      (+ sx (* sy 9))
					      nil)))
	       (tuitate-notation-setcar new
					tuitate-notation-data-turn
					turn))))
	  
	  ((string= (car elm) "DROP")
	   (let (dx dy piece check violation)
	     (if tuitate-notation-reverse
		 (setq dx (1- (string-to-number (nth 1 elm)))
		       dy (- 9 (string-to-number (nth 2 elm))))
	       (setq dx (- 9 (string-to-number (nth 1 elm)))
		     dy (1- (string-to-number (nth 2 elm)))))
	     (setq piece (cdr (assoc (nth 3 elm)
				     '(("PAWN"   . pawn)
				       ("LANCE"  . lance)
				       ("KNIGHT" . knight)
				       ("SILVER" . silver)
				       ("GOLD"   . gold)
				       ("BISHOP" . bishop)
				       ("ROOK"   . rook)))))
	     (setq check (string= (nth 4 elm) "CHECKED"))
	     (if (> (length elm) 5)
		 (setq violation (cdr (assoc (nth 5 elm)
					     '(("JUMPOVER" . jumpover)
					       ("CHECK"     . check)
					       ("DROPON"    . dropon)
					       ("CHECKMATEBYDROPPINGPAWN" .
						pawn))))))
	     (tuitate-notation-setcar new
				      tuitate-notation-data-violation
				      violation)
	     (tuitate-notation-setcar new
				      tuitate-notation-data-check
				      check)
	     (tuitate-notation-setcar new
				      tuitate-notation-data-play
				      (list 'drop dx dy piece))
	     (if violation
		 (tuitate-notation-setcar
		  new
		  (if left
		      tuitate-notation-data-left-life
		    tuitate-notation-data-right-life)
		  (1- (nth
		       (if left
			   tuitate-notation-data-left-life
			 tuitate-notation-data-right-life)
		       new))))
	     (let ((board (nth tuitate-notation-data-board new))
		   (left-belongings (nth
				     tuitate-notation-data-left-belongings
				     new))
		   (right-belongings (nth
				      tuitate-notation-data-right-belongings
				      new)))
	       (if (not violation)
		   (progn
		     (tuitate-notation-setcar
		      board
		      (+ dx (* dy 9))
		      (cdr (assq piece
				 (if left
				     '((pawn   . P)
				       (lance  . L)
				       (knight . N)
				       (silver . S)
				       (gold   . G)
				       (bishop . B)
				       (rook   . R))
				   '((pawn   . p)
				     (lance  . l)
				     (knight . n)
				     (silver . s)
				     (gold   . g)
				     (bishop . b)
				     (rook   . r))))))
		     (if left
			 (setcdr (assq piece left-belongings)
				 (1- (cdr (assq piece left-belongings))))
		       (setcdr (assq piece right-belongings)
			       (1- (cdr (assq piece right-belongings)))))))
	       (tuitate-notation-setcar
		new
		tuitate-notation-data-message
		(concat
		 (nth 1 elm)
		 (cdr (assoc (nth 2 elm)
			     '(("1" . "a")
			       ("2" . "b")
			       ("3" . "c")
			       ("4" . "d")
			       ("5" . "e")
			       ("6" . "f")
			       ("7" . "g")
			       ("8" . "h")
			       ("9" . "i"))))
		 " "
		 (cdr (assq piece
			    '((pawn   . "pawn")
			      (lance  . "lance")
			      (knight . "knight")
			      (silver . "silver")
			      (gold   . "gold")
			      (bishop . "bishop")
			      (rook   . "rook"))))
		 ". "
		 (if violation
		     (cdr (assq
			   violation
			   '((check . "Checked. ")
			     (dropon . "Dropping on a piece. ")
			     (pawn . "Checkmate by dropping a pawn. ")))
			  ))
		 (if check
		     "Check. ")))
	       (tuitate-notation-setcar new
					tuitate-notation-data-turn
					turn))))
	  
	  ((string= (car elm) "END")
	   (setq tuitate-notation-list nil) ; free list object
	   (let ((result (cdr (assoc (nth 1 elm)
				     '(("BLACK"  . black)
				       ("WHITE"  . white)
				       ("NOSIDE" . noside)))))
		 (reason (cdr (assoc (nth 2 elm)
				     '(("CHECKMATE" . checkmate)
				       ("LOSELIFE"  . loselife)
				       ("GIVEUP"    . giveup)
				       ("EMERGENCY" . emergency))))))
	     (tuitate-notation-setcar new
				      tuitate-notation-data-turn
				      turn)
	     (tuitate-notation-setcar new
				      tuitate-notation-data-play
				      (list 'end result reason))
	     (tuitate-notation-setcar new
				      tuitate-notation-data-message
				      (concat
				       (cdr (assq result
						  '((black . "Black win")
						    (white . "White win")
						    (noside . "Draw game"))
						  ))
				       " because of "
				       (cdr (assq
					     reason
					     '((checkmate . "checkmate")
					       (loselife . "losing life")
					       (giveup . "giving up")
					       (emergency . "emergency"))))
				       ))))
	  (t
	   (error "Invalid message")))
    (aset tuitate-notation-compiled-data
	  (setq tuitate-notation-last-data (1+ tuitate-notation-last-data))
	  new)))

;;;
;;; Commands.
;;;
(tuitate-notation-def-command
 tuitate-notation-redisplay ()
 "Re-display the buffer"
 (interactive)
 (let ((old-buffer (current-buffer))
       (old-marker (copy-marker (mark-marker))))
   (unwind-protect
       (progn
	 (set-buffer tuitate-buffer-name)
	 (let ((buffer-read-only nil))
	   (erase-buffer)
;;	   (tuitate-disp-insert-turn-string tuitate-notation-reverse)
	   (tuitate-disp-insert-name-string tuitate-notation-left-name
					    tuitate-notation-right-name)
	   (tuitate-disp-write-life-string 0 0) ; need not to be right
	   (tuitate-disp-insert-board tuitate-notation-reverse)
	   (tuitate-disp-write-belongings '((pawn   . 0)
					    (lance  . 0)
					    (knight . 0)
					    (silver . 0)
					    (gold   . 0)
					    (bishop . 0)
					    (rook   . 0))
					  '((pawn   . 0)
					    (lance  . 0)
					    (knight . 0)
					    (silver . 0)
					    (gold   . 0)
					    (bishop . 0)
					    (rook   . 0)))
	   (goto-char (point-max)))
	 (tuitate-notation-show))
     (set-buffer old-buffer)
     (set-marker (mark-marker) old-marker))))

;; Show current data.
(defun tuitate-notation-show ()
  (let* ((current (aref tuitate-notation-compiled-data
			tuitate-notation-current-data))
	 (turn (nth tuitate-notation-data-turn current))
	 (violation (nth tuitate-notation-data-violation current))
	 (check (nth tuitate-notation-data-check current))
	 (get (nth tuitate-notation-data-get current))
	 (promote (nth tuitate-notation-data-promote current))
	 (play (nth tuitate-notation-data-play current))
	 (left-life (nth tuitate-notation-data-left-life current))
	 (right-life (nth tuitate-notation-data-right-life current))	 
	 (board (nth tuitate-notation-data-board current))
	 (left-belongings (nth tuitate-notation-data-left-belongings
			       current))
	 (right-belongings (nth tuitate-notation-data-right-belongings
				current))
	 (msg (nth tuitate-notation-data-message current))
	 (left (or (and tuitate-notation-reverse
			(= (logand turn 1) 1))
		   (and (not tuitate-notation-reverse)
			(= (logand turn 1) 0)))))
    (if (or (= turn 0) (eq (car play) 'end))
	(setq tuitate-notation-turn-string "*")
      (setq tuitate-notation-turn-string (number-to-string turn))
      (tuitate-disp-turn left
			 tuitate-notation-reverse))
    (tuitate-disp-write-life-string left-life right-life)
    (tuitate-disp-write-belongings left-belongings right-belongings)
    (let ((x 0)
	  (y 0))
      (mapcar (function (lambda (item)
			  (tuitate-disp-write-piece
			   (cdr (assq item
				      '((p . pawn)
					(l . lance)
					(n . knight)
					(s . silver)
					(g . gold)
					(k . king)
					(b . bishop)
					(r . rook)
					(P . pawn)
					(L . lance)
					(N . knight)
					(S . silver)
					(G . gold)
					(K . king)
					(B . bishop)
					(R . rook)
					(p* . Pawn)
					(l* . Lance)
					(n* . Knight)
					(s* . Silver)
					(b* . Bishop)
					(r* . Rook)
					(P* . Pawn)
					(L* . Lance)
					(N* . Knight)
					(S* . Silver)
					(B* . Bishop)
					(R* . Rook))))
			   x
			   y
			   (memq item '(P L N S G K B R P* L* N* S* B* R*)))
			  (if (= x 8)
			      (setq x 0
				    y (1+ y))
			    (setq x (1+ x)))))
	      board))
    (cond ((eq (car-safe play) 'move)
	   (tuitate-disp-hold-piece (nth 1 play) (nth 2 play))
	   (tuitate-disp-candidate (nth 3 play) (nth 4 play)))
	  ((eq (car-safe play) 'drop)
	   (tuitate-disp-hold-belonging (cdr (assq (nth 3 play)
						   '((pawn   . 0)
						     (lance  . 1)
						     (knight . 2)
						     (silver . 3)
						     (gold   . 4)
						     (bishop . 5)
						     (rook   . 6))))
					nil
					left)
	   (tuitate-disp-candidate (nth 1 play) (nth 2 play)))
	  ((eq (car-safe play) 'end)
	   ())
	  (play
	   (error "Invalid compiled data")))
    (message msg)))

;; Hide current data.
(defun tuitate-notation-hide ()
  (let* ((current (aref tuitate-notation-compiled-data
			tuitate-notation-current-data))
	 (turn (nth tuitate-notation-data-turn current))
	 (play (nth tuitate-notation-data-play current))
	 (left (or (and tuitate-notation-reverse
			(= (logand turn 1) 1))
		   (and (not tuitate-notation-reverse)
			(= (logand turn 1) 0)))))
    (tuitate-disp-turn left
		       tuitate-notation-reverse
		       t)
    (cond ((eq (car-safe play) 'move)
	   (tuitate-disp-hold-piece (nth 1 play) (nth 2 play) t)
	   (tuitate-disp-candidate (nth 3 play) (nth 4 play) t))
	  ((eq (car-safe play) 'drop)
	   (tuitate-disp-hold-belonging (cdr (assq (nth 3 play)
						   '((pawn   . 0)
						     (lance  . 1)
						     (knight . 2)
						     (silver . 3)
						     (gold   . 4)
						     (bishop . 5)
						     (rook   . 6))))
					t
					left)
	   (tuitate-disp-candidate (nth 1 play) (nth 2 play) t))
	  ((eq (car-safe play) 'end)
	   ())
	  (play
	   (error "Invalid compiled data")))))

(tuitate-notation-def-command
 tuitate-notation-quit ()
 "Quit"
 (interactive)
 ;; free memory
 (setq tuitate-notation-compiled-data nil
       tuitate-notation-list nil)
 (kill-buffer tuitate-buffer-name))

(tuitate-notation-def-command
 tuitate-notation-forward-turn ()
 "Show forward turn"
 (interactive)
 (if (eq (car-safe (nth tuitate-notation-data-play
			(aref tuitate-notation-compiled-data
			      tuitate-notation-current-data)))
	 'end)
     (error "End of turn"))
 (if (= tuitate-notation-current-data tuitate-notation-last-data)
     (tuitate-notation-compile))
 (tuitate-notation-hide)
 (setq tuitate-notation-current-data (1+ tuitate-notation-current-data))
 (tuitate-notation-show))

(tuitate-notation-def-command
 tuitate-notation-backward-turn ()
 "Show backward turn"
 (interactive)
 (if (= tuitate-notation-current-data 0)
     (error "Beginning of turn"))
 (tuitate-notation-hide)
 (setq tuitate-notation-current-data (1- tuitate-notation-current-data))
 (tuitate-notation-show))

(tuitate-notation-def-command
 tuitate-notation-next-turn ()
 "Show forward valid turn"
 (interactive)
 (let ((current (aref tuitate-notation-compiled-data
		      tuitate-notation-current-data))
       (index tuitate-notation-current-data))
   (if (eq (car-safe (nth tuitate-notation-data-play current)) 'end)
       (error "End of turn"))
   (let ((turn (nth tuitate-notation-data-turn current)))
     (catch 'next-turn
       (while t
	 (if (= index tuitate-notation-last-data)
	     (tuitate-notation-compile))
	 (setq index (1+ index)
	       current (aref tuitate-notation-compiled-data index))
	 (if (or (and (< turn (nth tuitate-notation-data-turn current))
		      (null (nth tuitate-notation-data-violation current)))
		 (eq (car-safe (nth tuitate-notation-data-play current))
		     'end))
	     (throw 'next-turn nil)))))
   (tuitate-notation-hide)
   (setq tuitate-notation-current-data index)
   (tuitate-notation-show)))

(tuitate-notation-def-command
 tuitate-notation-previous-turn ()
 "Show backward valid turn"
 (interactive)
 (let ((current (aref tuitate-notation-compiled-data
		      tuitate-notation-current-data))
       (index tuitate-notation-current-data))
   (if (= tuitate-notation-current-data 0)
       (error "Beginning of turn"))
   (let ((turn (nth tuitate-notation-data-turn current)))
     (catch 'previous-turn
       (while t
	 (setq index (1- index)
	       current (aref tuitate-notation-compiled-data index))
	 (if (or (> turn (nth tuitate-notation-data-turn current))
		 (= index 0))
	     (throw 'previous-turn nil)))))
   (tuitate-notation-hide)
   (setq tuitate-notation-current-data index)
   (tuitate-notation-show)))

(tuitate-notation-def-command
 tuitate-notation-beginning-of-turn ()
 "Show the first turn"
 (interactive)
 (tuitate-notation-hide)
 (setq tuitate-notation-current-data 0)
 (tuitate-notation-show))

(tuitate-notation-def-command
 tuitate-notation-end-of-turn ()
 "Show the last turn"
 (interactive)
 ;; this may cost high.
 (while (not (eq (car-safe (nth tuitate-notation-data-play
				(aref tuitate-notation-compiled-data
				      tuitate-notation-last-data)))
		 'end))
   (tuitate-notation-compile))
 (tuitate-notation-hide)
 (setq tuitate-notation-current-data tuitate-notation-last-data)
 (tuitate-notation-show))

(tuitate-notation-def-command
 tuitate-notation-goto-turn (turn)
 "Show specified valid turn. Ask a turn number"
 (interactive "nTurn: ")
 (if (< turn 0)
     (error "Out of range"))
 (if (> turn (nth tuitate-notation-data-turn
		  (aref tuitate-notation-compiled-data
			tuitate-notation-last-data)))
     (while (not (eq (car-safe (nth tuitate-notation-data-play
				    (aref tuitate-notation-compiled-data
					  tuitate-notation-last-data)))
		     'end))
       (tuitate-notation-compile)))
 (if (> turn (nth tuitate-notation-data-turn
		  (aref tuitate-notation-compiled-data
			tuitate-notation-last-data)))
     (error "Out of range"))
 (let ((index turn))
   (while (or (> turn (nth tuitate-notation-data-turn
			   (aref tuitate-notation-compiled-data index)))
	      (nth tuitate-notation-data-violation
		   (aref tuitate-notation-compiled-data index)))
     (setq index (1+ index)))
   (tuitate-notation-hide)
   (setq tuitate-notation-current-data index)
   (tuitate-notation-show)))

(tuitate-notation-def-command
 tuitate-notation-re-message ()
 "Display last message again"
 (interactive)
 (message (nth tuitate-notation-data-message
	       (aref tuitate-notation-compiled-data
		     tuitate-notation-current-data))))

(tuitate-notation-def-command
 tuitate-notation-disp-information (&optional arg)
 "Display players' information"
 (interactive "P")
 (let* ((turn (if arg tuitate-white-string tuitate-black-string))
	(left (or (and arg (not tuitate-notation-reverse))
		  (and (not arg) tuitate-notation-reverse)))
	(name (if left
		  tuitate-notation-left-name
		tuitate-notation-right-name))
	(mail (if left
		  tuitate-notation-left-mail
		tuitate-notation-right-mail)))
   (message "%s: %s <%s>" turn name mail)))

;;;
;;; End of tui-not.el
;;;
