;;;			     tui-select.el
;;;
;;;	      Copyright (C) 1997       OKUJI Yoshinori
;;;
;;;		   This emacs lisp library conforms
;;;		GNU GENERAL PUBLIC LICENSE Version 2.
;;;
;;; Author:  OKUJI Yoshinori <okuji@kmc.kyoto-u.ac.jp>
;;; Created: May    15, 1997
;;; Revised: May    25, 1997
;;;

;;;
;;; Loading.
;;;
(provide 'tui-select)
(require 'tuitate)
(require 'tui-ipc)

;;;
;;; System variables.
;;;
(defvar tuitate-select-black-name nil)
(defvar tuitate-select-black-mail nil)
(defvar tuitate-select-white-name nil)
(defvar tuitate-select-white-mail nil)
(defvar tuitate-select-wait-flag nil)
(defvar tuitate-select-candidates nil)
(defvar tuitate-select-position nil)
(defvar tuitate-select-notation-list nil)
(defvar tuitate-select-opponent-name nil)
(defvar tuitate-select-opponent-mail nil)
(defvar tuitate-select-state nil)
(defvar tuitate-select-input nil)
(defvar tuitate-select-output nil)
(defvar tuitate-select-mode-map nil)
(if tuitate-select-mode-map nil
  (let ((map (make-keymap)))
    (suppress-keymap map)
    (define-key map "\ "    'tuitate-select-determine)
    (define-key map "\C-m"  'tuitate-select-determine)
    (define-key map "f"     'tuitate-select-determine)
    (define-key map "\C-f"  'tuitate-select-determine)
    (define-key map [right] 'tuitate-select-determine)
    (define-key map "n"     'tuitate-select-next-line)
    (define-key map "\C-n"  'tuitate-select-next-line)
    (define-key map [down]  'tuitate-select-next-line)
    (define-key map "p"     'tuitate-select-previous-line)
    (define-key map "\C-p"  'tuitate-select-previous-line)
    (define-key map [up]    'tuitate-select-previous-line)
    (define-key map "l"     'tuitate-select-update)
    (define-key map "\C-l"  'tuitate-select-update)
    (define-key map "q"     'tuitate-select-quit)
    (define-key map "<"     'tuitate-select-beginning-of-buffer)
    (define-key map "\M-<"  'tuitate-select-beginning-of-buffer)
    (define-key map ">"     'tuitate-select-end-of-buffer)
    (define-key map "\M->"  'tuitate-select-end-of-buffer)
    (define-key map "V"     'tuitate-version)
    
    (setq tuitate-select-mode-map map)))

;;;
;;; Constant variables.
;;;
(defconst tuitate-select-start-lines
  '((game . 3) (view . 4) (notation . 4)))
(defconst tuitate-select-line-tabs
  '((game . 1) (view . 4) (notation . 4)))
(defconst tuitate-select-start-col 8)
(defconst tuitate-select-point-col 6)

;;;
;;; Macros.
;;;

;; This macro avoids a bug that remains key bindings.
;; Maybe it is Emacs's bug.
(defmacro tuitate-select-def-command (name arglist doc interactive &rest body)
  (list 'defun name arglist doc interactive
	(nconc (list 'if 
		     (list 'not 
			   (list 'eq 
				 (list 'current-local-map)
				 (list 'eval
				       (list 'quote
					     'tuitate-select-mode-map))))
		     nil)
	       body)))

;;;
;;; Mode funtion.
;;;

;; tuitate-select-mode is special.
(put 'tuitate-select-mode 'mode-class 'special)

(defun tuitate-select-mode (&optional state)
  "Major mode for selection for tuitate shogi.
The first line shows useful information.
The others are selectable candidates.
\\{tuitate-select-mode-map}
"
  (interactive)
  (kill-all-local-variables)
  (switch-to-buffer tuitate-buffer-name)
  (setq major-mode 'tuitate-select-mode
	mode-name "tuitate-select"
	truncate-lines t
	buffer-read-only t)
  (use-local-map tuitate-select-mode-map)
  (tuitate-ipc-connect)
  (tuitate-ipc-set-filter 'tuitate-select-filter)
  (setq tuitate-ipc-receive-flag t)
  (let ((buffer-read-only nil))
    (erase-buffer))
  (setq tuitate-select-state state
	tuitate-select-candidates nil
	tuitate-select-position nil
	tuitate-select-notation-list nil
	tuitate-select-wait-flag t
	tuitate-select-input nil
	tuitate-select-output nil)
  (cond ((eq state 'game)
	 (tuitate-ipc-send-game tuitate-player-name tuitate-player-mail)
	 (tuitate-ipc-send-challengerlist)
	 (setq tuitate-select-output 'challengerlist))
	((eq state 'view)
	 (tuitate-ipc-send-view)
	 (tuitate-ipc-send-gamelist)
	 (setq tuitate-select-output 'gamelist))
	((eq state 'notation)
	 (tuitate-ipc-send-notation)
	 (tuitate-ipc-send-notationlist)
	 (setq tuitate-select-output 'notationlist))
	(t
	 (error "Unknown state.")))
  (run-hooks 'tuitate-select-mode-hook))

;;;
;;; Mode filter.
;;;
(defun tuitate-select-filter (proc string)
  (let ((msg (tuitate-ipc-receive string)))
    (while msg
      (cond ((eq tuitate-select-state 'game)
	     (tuitate-select-game-message msg))
	    ((eq tuitate-select-state 'view)
	     (tuitate-select-view-message msg))
	    ((eq tuitate-select-state 'notation)
	     (tuitate-select-notation-message msg))
	    (t
	     (error "Unknown state now.")))
      (setq msg (tuitate-ipc-receive "")))))

(defun tuitate-select-game-message (msg)
  (let ((name (car msg)))
    (cond ((string= name "CHALLENGERLIST")
	   (setq tuitate-select-input 'challengerlist
		 tuitate-select-candidates nil))
	  ((string= name "CHALLENGED")
	   (setq tuitate-select-input 'challenged)
	   (or tuitate-inhibit-beep (beep))
	   (if (y-or-n-p
		(concat (nth 2 msg) " <" (nth 3 msg)
			"> challenges you. Do you accept it? "))
	       (progn
		 (tuitate-ipc-send-accept)
		 (setq tuitate-select-output 'accept
		       tuitate-select-wait-flag t
		       tuitate-select-opponent-name (nth 2 msg)
		       tuitate-select-opponent-mail (nth 3 msg)))
	     (tuitate-ipc-send-refuse)
	     (setq tuitate-select-output 'refuse
		   tuitate-select-wait-flag nil)))
	  ((string= name "NOTEXIST")
	   (setq tuitate-select-input 'notexist
		 tuitate-select-wait-flag nil)
	   (cond ((eq tuitate-select-output 'challenge)
		  (or tuitate-inhibit-beep (beep))
		  (message "The challenged opponent %s <%s> disappears."
			   tuitate-select-opponent-name
			   tuitate-select-opponent-mail))
		 ((eq tuitate-select-output 'accept)
		  (or tuitate-inhibit-beep (beep))
		  (message "Your opponent %s <%s> disappears."
			   tuitate-select-opponent-name
			   tuitate-select-opponent-mail))
		 (t
		  (error "Why NOTEXIST message comes?"))))
	  ((string= name "BUSY")
	   (setq tuitate-select-input 'busy
		 tuitate-select-wait-flag nil)
	   (or tuitate-inhibit-beep (beep))
	   (message "The challenged opponent %s <%s> is busy now."
		    tuitate-select-opponent-name
		    tuitate-select-opponent-mail))
	  ((string= name "REFUSED")
	   (setq tuitate-select-input 'refused
		 tuitate-select-wait-flag nil)
	   (or tuitate-inhibit-beep (beep))
	   (message "%s <%s> refuses your challenge."
		    tuitate-select-opponent-name
		    tuitate-select-opponent-mail))
	  ((string= name "START")
	   (setq tuitate-select-input 'start
		 tuitate-ipc-receive-flag nil
		 tuitate-select-candidates nil) ; free memory
	   (require 'tui-game)
	   (tuitate-game-mode (string= (nth 1 msg) "BLACK")
			      tuitate-select-opponent-name
			      tuitate-select-opponent-mail))
	  (t
	   (cond ((eq tuitate-select-input 'challengerlist)
		  (if (string= name ".")
		      (progn
			(setq tuitate-select-input nil
			      tuitate-select-wait-flag nil)
			(let ((buffer-read-only nil)
			      (old-buffer (current-buffer))
			      (old-marker (copy-marker (mark-marker))))
			  (unwind-protect
			      (progn
				(set-buffer tuitate-buffer-name)
				(insert "Select your opponent. Now "
					(let ((len
					       (length
						tuitate-select-candidates)))
					  (cond ((= len 0)
						 (setq tuitate-select-position
						       nil)
						 "nobody exists.")
						((= len 1)
						 (setq tuitate-select-position
						       0)
						 "1 player exists.\n\n")
						(t
						 (setq tuitate-select-position
						       0)
						 (format
						  "%d players exist.\n\n"
						  len)))))
				(mapcar (function
					 (lambda (candidate)
					   (insert (make-string
						    tuitate-select-start-col
						    ?\ )
						   (nth 1 candidate)
						   " <"
						   (nth 2 candidate)
						   ">\n")))
					tuitate-select-candidates)
				(tuitate-select-goto-candidate
				 tuitate-select-position))
			    (set-buffer old-buffer)
			    (set-marker (mark-marker) old-marker))))
		    (setq tuitate-select-candidates
			  (append tuitate-select-candidates
				  (cons msg nil)))))
		 (t
		  (error "Unknown message.")))))))

(defun tuitate-select-view-message (msg)		 
  (let ((name (car msg)))
    (cond ((string= name "GAMELIST")
	   (setq tuitate-select-input 'gamelist
		 tuitate-select-candidates nil))
	  ((string= name "VIEWSTART")
	   (setq tuitate-select-input 'viewstart
		 tuitate-ipc-receive-flag nil
		 tuitate-select-candidates nil)	; free memory
	   (require 'tui-view)
	   (tuitate-view-mode tuitate-select-white-name
			      tuitate-select-white-mail
			      tuitate-select-black-name
			      tuitate-select-black-mail))
	  (t
	   (cond ((eq tuitate-select-input 'gamelist)
		  (if (string= name ".")
		      (progn
			(setq tuitate-select-input nil
			      tuitate-select-wait-flag nil)
			(let ((buffer-read-only nil)
			      (old-buffer (current-buffer))
			      (old-marker (copy-marker (mark-marker))))
			  (unwind-protect
			      (progn
				(set-buffer tuitate-buffer-name)
				(insert "Select a game. Now "
					(let ((len
					       (length
						tuitate-select-candidates)))
					  (cond ((= len 0)
						 (setq tuitate-select-position
						       nil)
						 "there is no game.")
						((= len 1)
						 (setq tuitate-select-position
						       0)
						 "there is 1 game.\n\n")
						(t
						 (setq tuitate-select-position
						       0)
						 (format
						  "there are %d games.\n\n"
						  len)))))
				(mapcar (function
					 (lambda (candidate)
					   (insert (make-string
						    tuitate-select-start-col
						    ?\ )
						   (nth 1 candidate)
						   " <"
						   (nth 2 candidate)
						   ">\n"
						   (make-string
						    (+ tuitate-select-start-col
						       8)
						    ?\ )
						   "vs.\n"
						   (make-string
						    tuitate-select-start-col
						    ?\ )
						   (nth 3 candidate)
						   " <"
						   (nth 4 candidate)
						   ">\n"
						   (make-string
						    tuitate-select-start-col
						    ?\ )
						   "-\n")))
					tuitate-select-candidates)
				(tuitate-select-goto-candidate
				 tuitate-select-position))
			    (set-buffer old-buffer)
			    (set-marker (mark-marker) old-marker))))
		    (setq tuitate-select-candidates
			  (append tuitate-select-candidates
				  (cons msg nil)))))
		 
		 (t
		  (error "Unknow message.")))))))

(defun tuitate-select-notation-message (msg)
  (let ((name (car msg)))
    (cond ((string= name "NOTATIONLIST")
	   (setq tuitate-select-input 'notationlist
		 tuitate-select-candidates nil))
	  ((string= name "NOTATION")
	   (setq tuitate-select-input 'notation))
	  (t
	   (cond ((eq tuitate-select-input 'notationlist)
		  (cond ((string= name ".")
			 (setq tuitate-select-input nil
			       tuitate-select-wait-flag nil)
			 (let ((buffer-read-only nil)
			       (old-buffer (current-buffer))
			       (old-marker (copy-marker (mark-marker))))
			   (unwind-protect
			       (progn
				 (set-buffer tuitate-buffer-name)
				 (insert "Select a notation. Now "
					 (let ((len
						(length
						 tuitate-select-candidates)))
					   (cond ((= len 0)
						  (setq tuitate-select-position
							nil)
						  "there is no notation.")
						 ((= len 1)
						  (setq tuitate-select-position
							0)
						  "there is 1 notation.\n\n")
						 (t
						  (setq tuitate-select-position
							0)
						  (format
						   (concat
						    "there are %d "
						    "notations.\n\n")
						   len)))))
				 (mapcar (function
					  (lambda (candidate)
					    (insert
					     (make-string
					      tuitate-select-start-col
					      ?\ )
					     (nth 3 candidate)
					     " <"
					     (nth 4 candidate)
					     ">\n"
					     (make-string
					      (+ tuitate-select-start-col 8)
					      ?\ )
					     "vs.\n"
					     (make-string
					      tuitate-select-start-col
					      ?\ )
					     (nth 5 candidate)
					     " <"
					     (nth 6 candidate)
					     ">\n"
					     (make-string
					      tuitate-select-start-col
					      ?\ )
					     "-\n")))
					 tuitate-select-candidates)
				 (tuitate-select-goto-candidate
				  tuitate-select-position))
			     (set-buffer old-buffer)
			     (set-marker (mark-marker) old-marker))))
			((and (>= (aref name 0) ?0) (<= (aref name 0) ?9))
			 (setq tuitate-select-candidates
			       (append tuitate-select-candidates
				       (cons msg nil))))
			(t
			 (let ((last (nth (1- (length
					       tuitate-select-candidates))
					  tuitate-select-candidates)))
			   (setcdr (nthcdr (1- (length last)) last)
				   msg)))))
		 ((eq tuitate-select-input 'notation)
		  (if (string= name ".")
		      (progn
			(setq tuitate-select-input nil
			      tuitate-select-wait-flag nil
			      tuitate-select-candidates nil) ; free memory
			(tuitate-ipc-shutdown)
			(require 'tui-not)
			(tuitate-notation-mode
			 tuitate-select-white-name tuitate-select-white-mail
			 tuitate-select-black-name tuitate-select-black-mail
			 tuitate-select-notation-list))
		    (setq tuitate-select-notation-list
			  (append tuitate-select-notation-list
				  (cons msg nil)))))
		 (t
		  (error "Unknown message.")))))))

;;;
;;; Commands.
;;;
(tuitate-select-def-command
 tuitate-select-determine ()
 "Select pointed candidate"
 (interactive)
 (if tuitate-select-wait-flag
     (error "Wait a while..."))
 (if tuitate-select-position
     (let ((result (nth tuitate-select-position tuitate-select-candidates)))
       (cond ((eq tuitate-select-state 'game)
	      (setq tuitate-select-opponent-name (nth 1 result)
		    tuitate-select-opponent-mail (nth 2 result))
	      (tuitate-ipc-send-challenge (string-to-number (nth 0 result)))
	      (setq tuitate-select-output 'challenge
		    tuitate-select-wait-flag t)
	      (message "Challenging %s <%s>..."
		       tuitate-select-opponent-name
		       tuitate-select-opponent-mail))
	     ((eq tuitate-select-state 'view)
	      (setq tuitate-select-black-name (nth 1 result)
		    tuitate-select-black-mail (nth 2 result)
		    tuitate-select-white-name (nth 3 result)
		    tuitate-select-white-mail (nth 4 result))
	      (tuitate-ipc-send-viewgame (string-to-number (nth 0 result)))
	      (setq tuitate-select-output 'viewgame
		    tuitate-select-wait-flag t)
	      (message "Trying to view %s vs. %s..."
		       tuitate-select-black-name
		       tuitate-select-white-name))
	     ((eq tuitate-select-state 'notation)
	      (setq tuitate-select-black-name (nth 3 result)
		    tuitate-select-black-mail (nth 4 result)
		    tuitate-select-white-name (nth 5 result)
		    tuitate-select-white-mail (nth 6 result))
	      (tuitate-ipc-send-viewnotation
	       (string-to-number (nth 0 result)))
	      (setq tuitate-select-output 'viewnotation
		    tuitate-select-wait-flag t)
	      (message "Trying to view %s vs. %s..."
		       tuitate-select-black-name
		       tuitate-select-white-name))))))

(tuitate-select-def-command
 tuitate-select-quit ()
 "Quit to select. "
 (interactive)
 (unwind-protect
     (progn
       (if (eq (process-status tuitate-ipc-process-name) 'open)
	   (progn
	     (tuitate-ipc-send-quit)
	     (setq tuitate-select-output 'quit)))
       (tuitate-ipc-shutdown))
   (setq tuitate-select-candidates nil) ; free memory
   (kill-buffer tuitate-buffer-name)))

(tuitate-select-def-command
 tuitate-select-update ()
 "Update information"
 (interactive)
 (if tuitate-select-wait-flag
     (error "Don\'t hurry"))
 (save-excursion
   (set-buffer tuitate-buffer-name)
   (let ((buffer-read-only nil))
     (erase-buffer)))
 (cond ((eq tuitate-select-state 'game)
	(tuitate-ipc-send-challengerlist)
	(setq tuitate-select-output 'challengerlist
	      tuitate-select-wait-flag t))
       ((eq tuitate-select-state 'view)
	(tuitate-ipc-send-gamelist)
	(setq tuitate-select-output 'gamelist
	      tuitate-select-wait-flag t))
       ((eq tuitate-select-state 'notation)
	(tuitate-ipc-send-notationlist)
	(setq tuitate-select-output 'notationlist
	      tuitate-select-wait-flag t))
       (t
	(error "Unknown state."))))

(defun tuitate-select-goto-candidate (position)
  (if position
      (progn
	(goto-line (+ (cdr (assq tuitate-select-state
				 tuitate-select-start-lines))
		      (* position (cdr (assq tuitate-select-state
					     tuitate-select-line-tabs)))))
	(forward-char tuitate-select-point-col))))

(tuitate-select-def-command
 tuitate-select-next-line ()
 "Go to next candidate"
 (interactive)
 (if tuitate-select-position
     (let ((old-buffer (current-buffer))
	   (old-marker (copy-marker (mark-marker))))
       (unwind-protect
	   (if (< tuitate-select-position
		  (1- (length tuitate-select-candidates)))
	       (progn
		 (setq tuitate-select-position (1+ tuitate-select-position))
		 (tuitate-select-goto-candidate tuitate-select-position)))
	 (set-buffer old-buffer)
	 (set-marker (mark-marker) old-marker)))))

(tuitate-select-def-command
 tuitate-select-previous-line ()
 "Go to previous candidate"
 (interactive)
 (if tuitate-select-position
     (let ((old-buffer (current-buffer))
	   (old-marker (copy-marker (mark-marker))))
       (unwind-protect
	   (if (> tuitate-select-position 0)
	       (progn
		 (setq tuitate-select-position (1- tuitate-select-position))
		 (tuitate-select-goto-candidate tuitate-select-position)))
	 (set-buffer old-buffer)
	 (set-marker (mark-marker) old-marker)))))

(tuitate-select-def-command
 tuitate-select-beginning-of-buffer ()
 "Go to the first candidate"
 (interactive)
 (if tuitate-select-position
     (let ((old-buffer (current-buffer))
	   (old-marker (copy-marker (mark-marker))))
       (unwind-protect
	   (progn
	     (setq tuitate-select-position 0)
	     (tuitate-select-goto-candidate 0))
	 (set-buffer old-buffer)
	 (set-marker (mark-marker) old-marker)))))

(tuitate-select-def-command
 tuitate-select-end-of-buffer ()
 "Go to the last candidate"
 (interactive)
 (if tuitate-select-position
     (let ((old-buffer (current-buffer))
	   (old-marker (copy-marker (mark-marker))))
       (unwind-protect
	   (progn
	     (setq tuitate-select-position
		   (1- (length tuitate-select-candidates)))
	     (tuitate-select-goto-candidate tuitate-select-position))
	 (set-buffer old-buffer)
	 (set-marker (mark-marker) old-marker)))))

;;;
;;; End of tui-select.el
;;;
