;;;			     tui-view.el
;;;
;;;	      Copyright (C) 1997       OKUJI Yoshinori
;;;
;;;		   This emacs lisp library conforms
;;;		GNU GENERAL PUBLIC LICENSE Version 2.
;;;
;;; Author:  OKUJI Yoshinori <okuji@kmc.kyoto-u.ac.jp>
;;; Created: May    15, 1997
;;; Revised: May    26, 1997
;;;

;;;
;;; Loading.
;;;
(provide 'tui-view)
(require 'tuitate)
(require 'tui-disp)
(require 'tui-ipc)

;;;
;;; System variables.
;;;
(defvar tuitate-view-current-data  0)
(defvar tuitate-view-last-data     0)
(defvar tuitate-view-turn-string   nil)
(defvar tuitate-view-left-name     nil)
(defvar tuitate-view-left-mail     nil)
(defvar tuitate-view-right-name    nil)
(defvar tuitate-view-right-mail    nil)
(defvar tuitate-view-compiled-data nil)
(defvar tuitate-view-mode-map      nil)
(if tuitate-view-mode-map nil
  (let ((map (make-keymap)))
    (suppress-keymap map)
    (define-key map "f"     'tuitate-view-forward-turn)
    (define-key map "\C-f"  'tuitate-view-forward-turn)
    (define-key map [right] 'tuitate-view-forward-turn)
    (define-key map "\ "    'tuitate-view-forward-turn)
    (define-key map "b"     'tuitate-view-backward-turn)
    (define-key map "\C-b"  'tuitate-view-backward-turn)
    (define-key map [left]  'tuitate-view-backward-turn)
    (define-key map "n"     'tuitate-view-next-turn)
    (define-key map "\C-n"  'tuitate-view-next-turn)
    (define-key map [down]  'tuitate-view-next-turn)
    (define-key map "\C-m"  'tuitate-view-next-turn)
    (define-key map "p"     'tuitate-view-previous-turn)
    (define-key map "\C-p"  'tuitate-view-previous-turn)
    (define-key map [up]    'tuitate-view-previous-turn)
    (define-key map "<"     'tuitate-view-beginning-of-turn)
    (define-key map "\M-<"  'tuitate-view-beginning-of-turn)
    (define-key map ">"     'tuitate-view-end-of-turn)
    (define-key map "\M->"  'tuitate-view-end-of-turn)
    (define-key map "l"     'tuitate-view-redisplay)
    (define-key map "\C-l"  'tuitate-view-redisplay)
    (define-key map "q"     'tuitate-view-quit)
    (define-key map "g"     'tuitate-view-goto-turn)
    (define-key map "."     'tuitate-view-re-message)
    (define-key map "v"     'tuitate-view-disp-information)
    (define-key map "V"     'tuitate-version)
    
    (setq tuitate-view-mode-map map)))

;;;
;;; Constant variables.
;;;

;; view mode compiles view list.
;; This compiled data is list, indexed below.
;; Each piece name is one character symbol, or followed by `*'(promoting).
(defconst tuitate-view-data-turn              0) ; turn number
(defconst tuitate-view-data-violation         1) ; what violation
(defconst tuitate-view-data-check             2) ; check flag
(defconst tuitate-view-data-get               3) ; which piece is gotten
(defconst tuitate-view-data-promote           4) ; promote flag
(defconst tuitate-view-data-play              5) ; play information
(defconst tuitate-view-data-left-life         6) ; left life
(defconst tuitate-view-data-right-life        7) ; right life
(defconst tuitate-view-data-board             8) ; board status
(defconst tuitate-view-data-left-belongings   9) ; left belongings
(defconst tuitate-view-data-right-belongings 10) ; right belongings
(defconst tuitate-view-data-message          11) ; message string

;; initial vetor length for compiled data
(defconst tuitate-view-initial-length 16)
;; additional vetor length for compiled data
(defconst tuitate-view-add-length     16)

;;;
;;; Macros.
;;;

;; I don't use cl.
(defmacro tuitate-view-setcar (list index obj)
  (list 'setcar (list 'nthcdr index list) obj))

;; This macro avoids a bug that remains key bindings.
;; Maybe it is Emacs's bug.
(defmacro tuitate-view-def-command (name arglist doc interactive &rest body)
  (list 'defun name arglist doc interactive
	(nconc (list 'if 
		     (list 'not 
			   (list 'eq 
				 (list 'current-local-map)
				 (list 'eval
				       (list 'quote
					     'tuitate-view-mode-map))))
		     nil)
	       body)))

;;;
;;; Mode function.
;;;

;; tuitate-view-mode is special.
(put 'tuitate-view-mode 'mode-class 'special)

(defun tuitate-view-mode (&optional left-name left-mail
				    right-name right-mail)
  "Major mode for viewing tuitate shogi.
The first line shows turn information.
The second line shows name information.
The third line shows life information.
The others are the board and the belongings.
\\{tuitate-view-mode-map}
"
  (interactive)
  (kill-all-local-variables)
  (switch-to-buffer tuitate-buffer-name)
  (setq major-mode 'tuitate-view-mode
	mode-name "tuitate-view"
	buffer-read-only t
	truncate-lines t)
  (if (equal (nth (- (length mode-line-format) 2) mode-line-format)
	     '(-3 . "%p"))
      (setq mode-line-format
	    (let ((mlf (copy-sequence mode-line-format))
		  (l (length mode-line-format)))
	      (setcdr (nthcdr (- l 3) mlf)
		      '("Turn " tuitate-view-turn-string "-%-"))
	      mlf)))
  (use-local-map tuitate-view-mode-map)
  (setq tuitate-view-turn-string "*"
	tuitate-view-left-name left-name
	tuitate-view-left-mail left-mail
	tuitate-view-right-name right-name
	tuitate-view-right-mail right-mail)
  (if (not (eq (process-status tuitate-ipc-process-name) 'open))
      (progn
	(kill-buffer tuitate-buffer-name)
	(error "Not connected yet.")))
  (tuitate-view-init-compile)
  (tuitate-disp-make-faces)
  (tuitate-view-redisplay)
  (tuitate-ipc-set-filter 'tuitate-view-filter)
  (setq tuitate-ipc-receive-flag t)
  (tuitate-view-filter nil "")		; deal with buffering messages.
  (run-hooks 'tuitate-view-mode-hook))

;; Initialize compile status.
(defun tuitate-view-init-compile ()
  (let (new)
    (setq tuitate-view-compiled-data (make-vector
				      tuitate-view-initial-length
				      nil)
	  tuitate-view-current-data 0
	  tuitate-view-last-data 0
	  new (list
	       0
	       nil
	       nil
	       nil
	       nil
	       nil
	       8
	       8
	       (list 'L  'N  'S  'G  'K  'G  'S  'N  'L
		     nil 'R  nil nil nil nil nil 'B  nil
		     'P  'P  'P  'P  'P  'P  'P  'P  'P
		     nil nil nil nil nil nil nil nil nil
		     nil nil nil nil nil nil nil nil nil
		     nil nil nil nil nil nil nil nil nil
		     'p  'p  'p  'p  'p  'p  'p  'p  'p
		     nil 'b  nil nil nil nil nil 'r  nil
		     'l  'n  's  'g  'k  'g  's  'n  'l )
	       (list (cons 'pawn 0) (cons 'lance 0) (cons 'knight 0)
		     (cons 'silver 0) (cons 'gold 0) (cons 'bishop 0)
		     (cons 'rook 0))
	       (list (cons 'pawn 0) (cons 'lance 0) (cons 'knight 0)
		     (cons 'silver 0) (cons 'gold 0) (cons 'bishop 0)
		     (cons 'rook 0))
	       "Game start!"))
    (aset tuitate-view-compiled-data 0 new)
    ))

;; Compile a message.
(defun tuitate-view-compile (msg)
  (let ((last (aref tuitate-view-compiled-data tuitate-view-last-data))
	new turn left)
    (setq new (list
	       nil
	       nil
	       nil
	       nil
	       nil
	       nil
	       (nth tuitate-view-data-left-life last)
	       (nth tuitate-view-data-right-life last)
	       (copy-sequence (nth tuitate-view-data-board last))
	       (copy-alist (nth tuitate-view-data-left-belongings last))
	       (copy-alist (nth tuitate-view-data-right-belongings last))
	       nil)
	  turn (nth tuitate-view-data-turn last))
    (if (null (nth tuitate-view-data-violation last))
	(setq turn (1+ turn)))
    (setq left (= (logand turn 1) 0))
    (cond ((string= (car msg) "MOVE")
	   (let ((sx (- 9 (string-to-number (nth 1 msg))))
		 (sy (1- (string-to-number (nth 2 msg))))
		 (dx (- 9 (string-to-number (nth 3 msg))))
		 (dy (1- (string-to-number (nth 4 msg))))
		 (promote (string= (nth 5 msg) "PROMOTE"))
		 (check (string= (nth 6 msg) "CHECKED"))
		 piece violation)
	     (if (> (length msg) 7)
		 (setq piece (cdr (assoc (nth 7 msg)
					 '(("PAWN"   . pawn)
					   ("LANCE"  . lance)
					   ("KNIGHT" . knight)
					   ("SILVER" . silver)
					   ("GOLD"   . gold)
					   ("BISHOP" . bishop)
					   ("ROOK"   . rook))))))
	     (if (> (length msg) 7)
		 (setq violation (cdr (assoc
				       (nth (1- (length msg)) msg)
				       '(("JUMPOVER" . jumpover)
					 ("CHECK"    . check)
					 ("DROPON"   . dropon)
					 ("CHECKMATEBYDROPPINGPAWN" . pawn))
				       ))))
	     (tuitate-view-setcar new
				  tuitate-view-data-violation
				  violation)
	     (tuitate-view-setcar new
				  tuitate-view-data-check
				  check)
	     (tuitate-view-setcar new
				  tuitate-view-data-get
				  piece)
	     (tuitate-view-setcar new
				  tuitate-view-data-promote
				  promote)
	     (tuitate-view-setcar new
				  tuitate-view-data-play
				  (list 'move sx sy dx dy))
	     (if violation
		 (tuitate-view-setcar
		  new
		  (if left
		      tuitate-view-data-left-life
		    tuitate-view-data-right-life)
		  (1- (nth
		       (if left
			   tuitate-view-data-left-life
			 tuitate-view-data-right-life)
		       new))))
	     (let ((board (nth tuitate-view-data-board new))
		   (left-belongings (nth
				     tuitate-view-data-left-belongings
				     new))
		   (right-belongings (nth
				      tuitate-view-data-right-belongings
				      new)))
	       (if piece
		   (if left
		       (setcdr (assq piece left-belongings)
			       (1+ (cdr (assq piece left-belongings))))
		     (setcdr (assq piece right-belongings)
			     (1+ (cdr (assq piece right-belongings))))))
	       (tuitate-view-setcar new
				    tuitate-view-data-message
				    (concat
				     (nth 3 msg)
				     (cdr (assoc (nth 4 msg)
						 '(("1" . "a")
						   ("2" . "b")
						   ("3" . "c")
						   ("4" . "d")
						   ("5" . "e")
						   ("6" . "f")
						   ("7" . "g")
						   ("8" . "h")
						   ("9" . "i"))))
				     " "
				     (cdr (assq
					   (nth
					    (+ sx (* sy 9))
					    board)
					   '((p . "pawn")
					     (l . "lance")
					     (n . "knight")
					     (s . "silver")
					     (g . "gold")
					     (k . "king")
					     (b . "bishop")
					     (r . "rook")
					     (P . "pawn")
					     (L . "lance")
					     (N . "knight")
					     (S . "silver")
					     (G . "gold")
					     (K . "king")
					     (B . "bishop")
					     (R . "rook")
					     (p* . "promoting pawn")
					     (l* . "promoting lance")
					     (n* . "promoting knight")
					     (s* . "promoting silver")
					     (b* . "promoting bishop")
					     (r* . "promoting rook")
					     (P* . "promoting pawn")
					     (L* . "promoting lance")
					     (N* . "promoting knight")
					     (S* . "promoting silver")
					     (B* . "promoting bishop")
					     (R* . "promoting rook"))))
				     (and promote " promote")
				     ". "
				     (if (and (not violation) piece)
					 (cdr (assq
					       piece
					       '((pawn . "Get pawn. ")
						 (lance . "Get lance. ")
						 (knight . "Get knight. ")
						 (silver . "Get silver. ")
						 (gold . "Get gold. ")
						 (bishop . "Get bishop. ")
						 (rook . "Get rook. ")))))
				     (if violation
					 (cdr
					  (assq
					   violation
					   '((jumpover . "Jumping over. ")
					     (check . "Checked. "))))
				       (if check
					   "Check. "))))
	       (if (not violation)
		   (progn
		     (tuitate-view-setcar
		      board
		      (+ dx (* dy 9))
		      (if promote
			  (cdr (assq (nth (+ sx (* sy 9))
					  board)
				     '((p . p*)
				       (l . l*)
				       (n . n*)
				       (s . s*)
				       (b . b*)
				       (r . r*)
				       (P . P*)
				       (L . L*)
				       (N . N*)
				       (S . S*)
				       (B . B*)
				       (R . R*))))
			(nth (+ sx (* sy 9)) board)))
		     (tuitate-view-setcar board
					  (+ sx (* sy 9))
					  nil)))
	       (tuitate-view-setcar new
				    tuitate-view-data-turn
				    turn))))
	  
	  ((string= (car msg) "DROP")
	   (let ((dx (- 9 (string-to-number (nth 1 msg))))
		 (dy (1- (string-to-number (nth 2 msg))))
		 (piece (cdr (assoc (nth 3 msg)
				    '(("PAWN"   . pawn)
				      ("LANCE"  . lance)
				      ("KNIGHT" . knight)
				      ("SILVER" . silver)
				      ("GOLD"   . gold)
				      ("BISHOP" . bishop)
				      ("ROOK"   . rook)))))
		 (check (string= (nth 4 msg) "CHECKED"))
		 violation)
	     (if (> (length msg) 5)
		 (setq violation (cdr (assoc
				       (nth 5 msg)
				       '(("JUMPOVER" . jumpover)
					 ("CHECK"     . check)
					 ("DROPON"    . dropon)
					 ("CHECKMATEBYDROPPINGPAWN" . pawn))
				       ))))
	     (tuitate-view-setcar new
				  tuitate-view-data-violation
				  violation)
	     (tuitate-view-setcar new
				  tuitate-view-data-check
				  check)
	     (tuitate-view-setcar new
				  tuitate-view-data-play
				  (list 'drop dx dy piece))
	     (if violation
		 (tuitate-view-setcar
		  new
		  (if left
		      tuitate-view-data-left-life
		    tuitate-view-data-right-life)
		  (1- (nth
		       (if left
			   tuitate-view-data-left-life
			 tuitate-view-data-right-life)
		       new))))
	     (let ((board (nth tuitate-view-data-board new))
		   (left-belongings (nth
				     tuitate-view-data-left-belongings
				     new))
		   (right-belongings (nth
				      tuitate-view-data-right-belongings
				      new)))
	       (if (not violation)
		   (progn
		     (tuitate-view-setcar
		      board
		      (+ dx (* dy 9))
		      (cdr (assq piece
				 (if left
				     '((pawn   . P)
				       (lance  . L)
				       (knight . N)
				       (silver . S)
				       (gold   . G)
				       (bishop . B)
				       (rook   . R))
				   '((pawn   . p)
				     (lance  . l)
				     (knight . n)
				     (silver . s)
				     (gold   . g)
				     (bishop . b)
				     (rook   . r))))))
		     (if left
			 (setcdr (assq piece left-belongings)
				 (1- (cdr (assq piece left-belongings))))
		       (setcdr (assq piece right-belongings)
			       (1- (cdr (assq piece right-belongings)))))))
	       (tuitate-view-setcar
		new
		tuitate-view-data-message
		(concat
		 (nth 1 msg)
		 (cdr (assoc (nth 2 msg)
			     '(("1" . "a")
			       ("2" . "b")
			       ("3" . "c")
			       ("4" . "d")
			       ("5" . "e")
			       ("6" . "f")
			       ("7" . "g")
			       ("8" . "h")
			       ("9" . "i"))))
		 " "
		 (cdr (assq piece
			    '((pawn   . "pawn")
			      (lance  . "lance")
			      (knight . "knight")
			      (silver . "silver")
			      (gold   . "gold")
			      (bishop . "bishop")
			      (rook   . "rook"))))
		 ". "
		 (if violation
		     (cdr
		      (assq
		       violation
		       '((check . "Checked. ")
			 (dropon . "Dropping on a piece. ")
			 (pawn . "Checkmate by dropping a pawn. ")))
		      ))
		 (if check
		     "Check. "))))
	     (tuitate-view-setcar new
				  tuitate-view-data-turn
				  turn)))
	  
	  ((string= (car msg) "END")
	   (let ((result (cdr (assoc (nth 1 msg)
				     '(("BLACK"  . black)
				       ("WHITE"  . white)
				       ("NOSIDE" . noside)))))
		 (reason (cdr (assoc (nth 2 msg)
				     '(("CHECKMATE" . checkmate)
				       ("LOSELIFE"  . loselife)
				       ("GIVEUP"    . giveup)
				       ("EMERGENCY" . emergency))))))
	     (tuitate-view-setcar new
				  tuitate-view-data-turn
				  turn)
	     (tuitate-view-setcar new
				  tuitate-view-data-play
				  (list 'end result reason))
	     (tuitate-view-setcar new
				  tuitate-view-data-message
				  (concat
				   (cdr (assq result
					      '((black . "Black win")
						(white . "White win")
						(noside . "Draw game"))
					      ))
				   " because of "
				   (cdr (assq
					 reason
					 '((checkmate . "checkmate")
					   (loselife . "losing life")
					   (giveup . "giving up")
					   (emergency . "emergency"))))
				   ))))
	  ((null msg)
	   (error "Next turn have not been played yet!"))
	  (t
	   (error "Invalid message")))
    (if (= (1- (length tuitate-view-compiled-data)) tuitate-view-last-data)
	(setq tuitate-view-compiled-data (vconcat
					  tuitate-view-compiled-data
					  (make-vector tuitate-view-add-length
						       nil))))
    (aset tuitate-view-compiled-data
	  (setq tuitate-view-last-data (1+ tuitate-view-last-data))
	  new)))

;;;
;;; Commands.
;;;
(tuitate-view-def-command
 tuitate-view-redisplay ()
 "Re-display the buffer"
 (interactive)
 (let ((old-buffer (current-buffer))
       (old-marker (copy-marker (mark-marker))))
   (unwind-protect
       (progn
	 (set-buffer tuitate-buffer-name)
	 (let ((buffer-read-only nil))
	   (erase-buffer)
;;	   (tuitate-disp-insert-turn-string)
	   (tuitate-disp-insert-name-string tuitate-view-left-name
					    tuitate-view-right-name)
	   (tuitate-disp-write-life-string 0 0) ; need not to be right
	   (tuitate-disp-insert-board)
	   (tuitate-disp-write-belongings '((pawn   . 0)
					    (lance  . 0)
					    (knight . 0)
					    (silver . 0)
					    (gold   . 0)
					    (bishop . 0)
					    (rook   . 0))
					  '((pawn   . 0)
					    (lance  . 0)
					    (knight . 0)
					    (silver . 0)
					    (gold   . 0)
					    (bishop . 0)
					    (rook   . 0)))
	   (goto-char (point-max)))
	 (tuitate-view-show))
     (set-buffer old-buffer)
     (set-marker (mark-marker) old-marker))))

;; Show current data.
(defun tuitate-view-show ()
  (let* ((current (aref tuitate-view-compiled-data
			tuitate-view-current-data))
	 (turn (nth tuitate-view-data-turn current))
	 (violation (nth tuitate-view-data-violation current))
	 (check (nth tuitate-view-data-check current))
	 (get (nth tuitate-view-data-get current))
	 (promote (nth tuitate-view-data-promote current))
	 (play (nth tuitate-view-data-play current))
	 (left-life (nth tuitate-view-data-left-life current))
	 (right-life (nth tuitate-view-data-right-life current))	 
	 (board (nth tuitate-view-data-board current))
	 (left-belongings (nth tuitate-view-data-left-belongings
			       current))
	 (right-belongings (nth tuitate-view-data-right-belongings
				current))
	 (msg (nth tuitate-view-data-message current))
	 (left (= (logand turn 1) 0)))
    (if (or (= turn 0) (eq (car play) 'end))
	(setq tuitate-view-turn-string "*")
      (setq tuitate-view-turn-string (number-to-string turn))
      (tuitate-disp-turn left))
    (tuitate-disp-write-life-string left-life right-life)
    (tuitate-disp-write-belongings left-belongings right-belongings)
    (let ((x 0)
	  (y 0))
      (mapcar (function (lambda (item)
			  (tuitate-disp-write-piece
			   (cdr (assq item
				      '((p . pawn)
					(l . lance)
					(n . knight)
					(s . silver)
					(g . gold)
					(k . king)
					(b . bishop)
					(r . rook)
					(P . pawn)
					(L . lance)
					(N . knight)
					(S . silver)
					(G . gold)
					(K . king)
					(B . bishop)
					(R . rook)
					(p* . Pawn)
					(l* . Lance)
					(n* . Knight)
					(s* . Silver)
					(b* . Bishop)
					(r* . Rook)
					(P* . Pawn)
					(L* . Lance)
					(N* . Knight)
					(S* . Silver)
					(B* . Bishop)
					(R* . Rook))))
			   x
			   y
			   (memq item '(P L N S G K B R P* L* N* S* B* R*)))
			  (if (= x 8)
			      (setq x 0
				    y (1+ y))
			    (setq x (1+ x)))))
	      board))
    (cond ((eq (car-safe play) 'move)
	   (tuitate-disp-hold-piece (nth 1 play) (nth 2 play))
	   (tuitate-disp-candidate (nth 3 play) (nth 4 play)))
	  ((eq (car-safe play) 'drop)
	   (tuitate-disp-hold-belonging (cdr (assq (nth 3 play)
						   '((pawn   . 0)
						     (lance  . 1)
						     (knight . 2)
						     (silver . 3)
						     (gold   . 4)
						     (bishop . 5)
						     (rook   . 6))))
					nil
					left)
	   (tuitate-disp-candidate (nth 1 play) (nth 2 play)))
	  ((eq (car-safe play) 'end)
	   ())
	  (play
	   (error "Invalid compiled data")))
    (message msg)))

;; Hide current data.
(defun tuitate-view-hide ()
  (let* ((current (aref tuitate-view-compiled-data
			tuitate-view-current-data))
	 (turn (nth tuitate-view-data-turn current))
	 (play (nth tuitate-view-data-play current))
	 (left (= (logand turn 1) 0)))
    (tuitate-disp-turn left
		       nil
		       t)
    (cond ((eq (car-safe play) 'move)
	   (tuitate-disp-hold-piece (nth 1 play) (nth 2 play) t)
	   (tuitate-disp-candidate (nth 3 play) (nth 4 play) t))
	  ((eq (car-safe play) 'drop)
	   (tuitate-disp-hold-belonging (cdr (assq (nth 3 play)
						   '((pawn   . 0)
						     (lance  . 1)
						     (knight . 2)
						     (silver . 3)
						     (gold   . 4)
						     (bishop . 5)
						     (rook   . 6))))
					t
					left)
	   (tuitate-disp-candidate (nth 1 play) (nth 2 play) t))
	  ((eq (car-safe play) 'end)
	   ())
	  (play
	   (error "Invalid compiled data")))))

(tuitate-view-def-command
 tuitate-view-quit ()
 "Quit"
 (interactive)
 (unwind-protect
     (progn
       (if (eq (process-status tuitate-ipc-process-name) 'open)
	   (tuitate-ipc-send-quit))
       (tuitate-ipc-shutdown))
   (setq tuitate-view-compiled-data nil) ; free memory
   (kill-buffer tuitate-buffer-name)))

(tuitate-view-def-command
 tuitate-view-forward-turn ()
 "Show forward turn"
 (interactive)
 (if (= tuitate-view-current-data tuitate-view-last-data)
     (error "End of turn"))
 (tuitate-view-hide)
 (setq tuitate-view-current-data (1+ tuitate-view-current-data))
 (tuitate-view-show))

(tuitate-view-def-command
 tuitate-view-backward-turn ()
 "Show backward turn"
 (interactive)
 (if (= tuitate-view-current-data 0)
     (error "Beginning of turn"))
 (tuitate-view-hide)
 (setq tuitate-view-current-data (1- tuitate-view-current-data))
 (tuitate-view-show))

(tuitate-view-def-command
 tuitate-view-next-turn ()
 "Show forward valid turn"
 (interactive)
 (let ((current (aref tuitate-view-compiled-data
		      tuitate-view-current-data))
       (index tuitate-view-current-data))
   (if (= tuitate-view-current-data tuitate-view-last-data)
       (error "End of turn"))
   (let ((turn (nth tuitate-view-data-turn current)))
     (catch 'next-turn
       (while t
	 (setq index (1+ index)
	       current (aref tuitate-view-compiled-data index))
	 (if (or (and (< turn (nth tuitate-view-data-turn current))
		      (null (nth tuitate-view-data-violation current)))
		 (= index tuitate-view-last-data))
	     (throw 'next-turn nil)))))
   (tuitate-view-hide)
   (setq tuitate-view-current-data index)
   (tuitate-view-show)))

(tuitate-view-def-command
 tuitate-view-previous-turn ()
 "Show backward valid turn"
 (interactive)
 (let ((current (aref tuitate-view-compiled-data
		      tuitate-view-current-data))
       (index tuitate-view-current-data))
   (if (= tuitate-view-current-data 0)
       (error "Beginning of turn"))
   (let ((turn (nth tuitate-view-data-turn current)))
     (catch 'previous-turn
       (while t
	 (setq index (1- index)
	       current (aref tuitate-view-compiled-data index))
	 (if (or (> turn (nth tuitate-view-data-turn current))
		 (= index 0))
	     (throw 'previous-turn nil)))))
   (tuitate-view-hide)
   (setq tuitate-view-current-data index)
   (tuitate-view-show)))

(tuitate-view-def-command
 tuitate-view-beginning-of-turn ()
 "Show the first turn"
 (interactive)
 (tuitate-view-hide)
 (setq tuitate-view-current-data 0)
 (tuitate-view-show))

(tuitate-view-def-command
 tuitate-view-end-of-turn ()
 "Show the last turn"
 (interactive)
 (tuitate-view-hide)
 (setq tuitate-view-current-data tuitate-view-last-data)
 (tuitate-view-show))

(tuitate-view-def-command
 tuitate-view-goto-turn (turn)
 "Show specified valid turn. Ask a turn number"
 (interactive "nTurn: ")
 (if (< turn 0)
     (error "Out of range"))
 (if (> turn (nth tuitate-view-data-turn
		  (aref tuitate-view-compiled-data
			tuitate-view-last-data)))
     (error "Out of range"))
 (let ((index turn))
   (while (or (> turn (nth tuitate-view-data-turn
			   (aref tuitate-view-compiled-data index)))
	      (nth tuitate-view-data-violation
		   (aref tuitate-view-compiled-data index)))
     (setq index (1+ index)))
   (tuitate-view-hide)
   (setq tuitate-view-current-data index)
   (tuitate-view-show)))

(tuitate-view-def-command
 tuitate-view-re-message ()
 "Display last message again"
 (interactive)
 (message (nth tuitate-view-data-message
	       (aref tuitate-view-compiled-data
		     tuitate-view-current-data))))

(tuitate-view-def-command
 tuitate-view-disp-information (&optional arg)
 "Display players' information"
 (interactive "P")
 (if arg
     (message "%s: %s <%s>"
	      tuitate-white-string
	      tuitate-view-left-name
	      tuitate-view-left-mail)
   (message "%s: %s <%s>"
	    tuitate-black-string
	    tuitate-view-right-name
	    tuitate-view-right-mail)))

;;;
;;; Mode filter.
;;;
(defun tuitate-view-filter (proc string)
  (let ((msg (tuitate-ipc-receive string)))
    (while msg
      (tuitate-view-compile msg)
      (if (string= (car msg) "END")
	  (tuitate-ipc-shutdown))
      (setq msg (tuitate-ipc-receive ""))))
  (if tuitate-follow-flag
      (progn
	(tuitate-view-hide)
	(setq tuitate-view-current-data tuitate-view-last-data)
	(or tuitate-inhibit-beep (beep))
	(tuitate-view-show))))

;;;
;;; End of tui-view.el
;;;
