;;;	        tuitate.el :: tuitate client for Emacs
;;;
;;;	      Copyright (C) 1997      OKUJI Yoshinori
;;;
;;;		   This emacs lisp library conforms
;;;		GNU GENERAL PUBLIC LICENSE Version 2.
;;;
;;; Author:  OKUJI Yoshinori <okuji@kmc.kyoto-u.ac.jp>
;;; Created: May   15, 1997
;;; Revised: Jul    5, 1997
;;;
;;; The updated version is available from:
;;;     http://www.kuis.kyoto-u.ac.jp/~kmc/proj/tuitate/emacs/index.html
;;;
;;; Minimum setup:
;;;	(autoload 'tuitate          "tuitate" "Tuitate shogi game!"     t)
;;;	(autoload 'tuitate-view     "tuitate" "Tuitate shogi view!"     t)
;;;	(autoload 'tuitate-notation "tuitate" "Tuitate shogi notation!" t)

;;;
;;; Loading.
;;;
(provide 'tuitate)
(require 'tui-select)

;;;
;;; User customozable variables.
;;;
(defvar tuitate-player-name (user-login-name)
  "*Your name for tuitate shogi")
(defvar tuitate-player-mail user-mail-address
  "*Your e-mail address for tuitate shogi")
(defvar tuitate-server "localhost"
  "*A host that runs the tuitate server(tuitated)")
(defvar tuitate-port 9070
  "*The port number that the tuitate server(tuitated) use")
(defvar tuitate-black-string "Black"
  "*The string for Black")
(defvar tuitate-white-string "White"
  "*The string for White")
(defvar tuitate-you-string "You"
  "*The string for you")
(defvar tuitate-opponent-string "Opponent"
  "*The string for your opponent")
(defvar tuitate-name-string "Name"
  "*The string for players' name")
(defvar tuitate-life-string "Life"
  "*The string for life")
(defvar tuitate-follow-flag t
  "*If non-nil, you follow that the viewing game proceeds")
(defvar tuitate-inhibit-color nil
  "*If nil, you use color faces")
(defvar tuitate-inhibit-beep t
  "*If nil, you ring a beep when something important happens")
(defvar tuitate-inhibit-terminal-face t
  "*If nil, you use terminal faces")

;;;
;;; Hooks.
;;;
(defvar tuitate-init-hook          nil
  "*Hook called at initalize time")
(defvar tuitate-select-mode-hook   nil
  "*Hook called in select mode")
(defvar tuitate-game-mode-hook     nil
  "*Hook called in game mode")
(defvar tuitate-view-mode-hook     nil
  "*Hook called in view mode")
(defvar tuitate-notation-mode-hook nil
  "*Hook called in notation mode")

;;;
;;; Constant variables.
;;;
(defconst tuitate-version     "4.0")
(defconst tuitate-buffer-name "*Tuitate*")

;;;
;;; Make to run GNU Emacs 19.28.1, and older ones.
;;;
(if (and (string-match "^[0-9]+\\.\\([0-9]+\\)" emacs-version)
	 (< (string-to-int (substring emacs-version
				      (match-beginning 1)
				      (match-end 1)))
	    29))
    (progn
      (defun rassoc (key list)
	(catch 'rassoc
	  (mapcar (function (lambda (item)
			      (if (equal key (cdr item))
				  (throw 'rassoc item))))
		  list)
	  nil))))

;;;
;;; Create the tuitate buffer.
;;;
(defun tuitate-create-buffer ()
  (let ((buffer (get-buffer tuitate-buffer-name)))
    (if buffer
	(save-excursion
	  (set-buffer buffer)
	  (let ((buffer-read-only nil))
	    (erase-buffer)))
      (generate-new-buffer tuitate-buffer-name))
    (switch-to-buffer tuitate-buffer-name)))

;;;
;;; Commands that user starts the emacs-tuitate.
;;;
(defun tuitate (&optional arg)
  "Entry tuitate shogi game"
  (interactive "P")
  (tuitate-create-buffer)
  (run-hooks 'tuitate-init-hook)
  (if arg
      (let ((server (read-string (format "The tuitate server(%s): "
					 tuitate-server))))
	(if server
	    (setq tuitate-server server))))
  (tuitate-select-mode 'game))

(defun tuitate-view (&optional arg)
  "View tuitate shogi game"
  (interactive "P")
  (tuitate-create-buffer)
  (run-hooks 'tuitate-init-hook)
  (if arg
      (let ((server (read-string (format "The tuitate server(%s): "
					 tuitate-server))))
	(if server
	    (setq tuitate-server server))))
  (tuitate-select-mode 'view))

(defun tuitate-notation (&optional arg)
  "View tuitate shogi game notation"
  (interactive "P")
  (tuitate-create-buffer)
  (run-hooks 'tuitate-init-hook)
  (if arg
      (let ((server (read-string (format "The tuitate server(%s): "
					 tuitate-server))))
	(if server
	    (setq tuitate-server server))))
  (tuitate-select-mode 'notation))

;; Display the version.
(defun tuitate-version ()
  "Show emacs-tuitate version"
  (interactive)
  (message "emacs-tuitate version %s, (C) OKUJI Yoshinori"
	   tuitate-version))

;;;
;;; End of tuitate.el
;;;
